#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import subprocess

import pytest
from sos import execute_workflow
from sos.converter import extract_workflow
from sos.parser import ParsingError, SoS_Script
from sos.targets import file_target, path, paths, sos_targets
from sos.utils import ArgumentError, env
from sos.workflow_executor import Base_Executor

section1_sos = """
#!/usr/bin/env sos-runner
#fileformat=SOS1.1

#
# this is a test sos script
#
var1='value1'
var2 = 'value2'
var3 = [var1,
  var2]

# section
# description for workflow section
#
[parameters]
# par1 string
par1 = 'var1'

# par2 list
par2 = ['a', 'b', 'c']

# par3 multiline
par3 = ['a',
	'b']

[*_0]
var0 = '0'

[section_10]
#
#step 10
var1 = 'a'

[section_2 : shared='var3']
var2 = 'a'
input: var1
output: var2

var3 = 'a'

[section_3, *_4 : shared='var4']
output:
	var2,
	var3

print()
var4= 'value4'

[chapter_5]
var4='5'
 """


def test_file_format():
    """Test recognizing the format of SoS script"""
    # file format must be 'fileformat=SOSx.x'
    with pytest.raises(ParsingError):
        SoS_Script("#fileformat=SS2")
    with pytest.raises(ParsingError):
        SoS_Script("#fileformat=SOS1.0beta")
    #
    # parse a larger script with gormat 1.1
    SoS_Script(section1_sos)
    # not the default value of 1.0
    # assert script.format_version, '1.1')


def test_workflows():
    """Test workflows defined in SoS script"""
    script = SoS_Script("""[0]""")
    assert sorted(script.workflows) == [""]
    script = SoS_Script("""[0]\n[1]""")
    assert sorted(script.workflows) == [""]
    script = SoS_Script("""[0]\n[*_1]""")
    assert sorted(script.workflows) == [""]
    script = SoS_Script("""[0]\n[*_1]\n[auxiliary:provides='{a}.txt']""")
    assert sorted(script.workflows) == ["", "auxiliary"]
    script = SoS_Script("""[0]\n[*_1]\n[human_1]""")
    assert sorted(script.workflows) == ["", "human"]
    script = SoS_Script("""[0]\n[*_1]\n[human_1]\n[mouse_2]""")
    assert sorted(script.workflows) == ["", "human", "mouse"]
    script = SoS_Script("""[0]\n[*_1]\n[human_1]\n[mouse_2]\n[s*_2]""")
    assert sorted(script.workflows) == ["", "human", "mouse"]
    # unnamed
    script = SoS_Script("""[0]\n[*_1]\n[human_1]\n[mouse]\n[s*_2]""")
    assert sorted(script.workflows) == ["", "human", "mouse"]
    #
    # workflow name with -
    script = SoS_Script("""[proc-1]\n[test-case_2]""")
    assert sorted(script.workflows) == ["proc-1", "test-case"]
    script.workflow("proc-1")
    script.workflow("proc-1 + test-case:2")


def test_sections():
    """Test section definitions"""
    # bad names
    for badname in ["56_1", "_a", "a_", "1x", "*", "?"]:
        with pytest.raises(ParsingError):
            SoS_Script("[{}]".format(badname))
    # bad options
    for badoption in ["ss"]:
        with pytest.raises(ParsingError):
            SoS_Script("[0:{}]".format(badoption))
    # allowed names
    for name in ["a5", "a_5", "*_0", "a*1_100"]:
        SoS_Script("[{}]".format(name))
    # allowed names with alias
    for name in ["a5 (p1)", "a_5 (something fun)", "*_0 (no way)", "a*1_100"]:
        SoS_Script("[{}]".format(name))
    # duplicate sections
    with pytest.raises(ParsingError):
        SoS_Script("""[1]\n[1]""")
    with pytest.raises(ParsingError):
        SoS_Script("""[1]\n[3]\n[2,1]""")
    with pytest.raises(ParsingError):
        SoS_Script("""[a_1]\n[a_3]\n[*_1]""")
    #
    # no duplicated section header
    SoS_Script("""[a_1]\n[a_3]\n[b*_1]""")
    #
    # global section
    with pytest.raises(ParsingError):
        SoS_Script("""[global, step_10]""")


def test_parameters():
    """Test ending parameters with new line #1311"""
    with pytest.raises(ParsingError):
        SoS_Script("""input: ['a.txt'\n\n'b.txt']\n""")


def test_global_variables():
    """Test definition of variables"""
    # allow definition
    SoS_Script("""a = '1' """)
    SoS_Script("""a = ['a', 'b'] """)
    # but this one has incorrect syntax
    with pytest.raises(ParsingError):
        SoS_Script("""a = 'b  """)

    SoS_Script('''a = """
this is a multi line
string """
''')
    # multi-line list literal, even with newline in between
    SoS_Script("""a = [
'a',

'b'
]
""")
    #
    SoS_Script(section1_sos)
    # not the default value of 1.0
    #
    script = SoS_Script("""
[global]
a = 1

[b]
print(a)
""")
    wf = script.workflow()
    Base_Executor(wf).run()


def test_parameters_section():
    """Test parameters section"""
    # directive not allowed in parameters
    script = SoS_Script(section1_sos)
    wf = script.workflow("chapter:0")

    #
    script = SoS_Script("""
parameter: a = [1, 2]
[0]
""")
    wf = script.workflow()
    assert list(wf.parameters().keys()) == ["a"]
    Base_Executor(wf).run()
    assert env.sos_dict["a"] == [1, 2]
    env.sos_dict.pop("a")
    wf = script.workflow()
    Base_Executor(wf, args=["--a", "3"]).run()
    assert env.sos_dict["a"] == [3]
    env.sos_dict.pop("a")
    wf = script.workflow()
    Base_Executor(wf, args=["--a", "3", "5"]).run()
    assert env.sos_dict["a"] == [3, 5]
    env.sos_dict.pop("a")
    #
    script = SoS_Script("""
# comment
# comment
parameter: a = ['a.txt', 'b.txt']
[0]
""")
    wf = script.workflow()
    Base_Executor(wf).run()
    assert env.sos_dict["a"] == ["a.txt", "b.txt"]
    env.sos_dict.pop("a")
    wf = script.workflow()
    Base_Executor(wf, args=["--a", "3"]).run()
    assert env.sos_dict["a"] == ["3"]
    env.sos_dict.pop("a")
    wf = script.workflow()
    Base_Executor(wf, args=["--a", "3", "5"]).run()
    assert env.sos_dict["a"] == ["3", "5"]
    env.sos_dict.pop("a")
    #
    # test parameter using global definition
    script = SoS_Script("""
a="100"

# comment
parameter: b=str(int(a)+1)
[0]
""")
    wf = script.workflow()
    assert list(wf.parameters().keys()) == ["b"]
    Base_Executor(wf).run()
    assert env.sos_dict["b"] == "101"
    env.sos_dict.pop("b")
    #
    env.sos_dict.clear()
    script = SoS_Script("""
a=100

parameter: b=a+1
[0]
""")
    wf = script.workflow()
    Base_Executor(wf).run()
    assert env.sos_dict["b"] == 101
    env.sos_dict.pop("b")
    #
    script = SoS_Script("""
a=100

parameter: b=a+1.
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "1000"]).run()
    #
    assert env.sos_dict["b"] == 1000
    env.sos_dict.pop("b")
    #
    # argument has hve a value
    with pytest.raises(ParsingError):
        SoS_Script("""
[0]
parameter: b=

""")
    # if it is a type, must provide value
    script = SoS_Script("""
# comment
parameter: b = int
[0]
""")
    wf = script.workflow()
    with pytest.raises(ArgumentError):
        Base_Executor(wf)
    #
    script = SoS_Script("""
parameter: b = list
[0]
""")
    assert list(wf.parameters().keys()) == ["b"]
    wf = script.workflow()
    with pytest.raises(ArgumentError):
        Base_Executor(wf)
    # also require the type
    script = SoS_Script("""
parameter: b = int
[0]
""")
    wf = script.workflow()
    with pytest.raises(ArgumentError):
        Base_Executor(wf, args=["--b", "file"])
    #
    script = SoS_Script("""
parameter: b = int
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
    assert env.sos_dict["b"] == 5
    env.sos_dict.pop("b")
    # string
    script = SoS_Script("""
parameter: b = str
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
    assert env.sos_dict["b"] == "5"
    env.sos_dict.pop("b")
    # list is ok
    script = SoS_Script("""
parameter: b = list
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
    assert env.sos_dict["b"] == ["5"]
    # bool required
    script = SoS_Script("""
# comment
parameter: b = bool
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b"]).run(mode="dryrun")
    assert env.sos_dict["b"] == True
    env.sos_dict.pop("b")
    Base_Executor(wf, args=["--no-b"]).run(mode="dryrun")
    assert env.sos_dict["b"] == False
    env.sos_dict.pop("b")
    # bool with default True
    script = SoS_Script("""
parameter: b = True
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=[]).run(mode="dryrun")
    assert env.sos_dict["b"] == True
    env.sos_dict.pop("b")
    Base_Executor(wf, args=["--b"]).run(mode="dryrun")
    assert env.sos_dict["b"] == True
    env.sos_dict.pop("b")
    Base_Executor(wf, args=["--no-b"]).run(mode="dryrun")
    assert env.sos_dict["b"] == False
    env.sos_dict.pop("b")
    # bool with default False
    script = SoS_Script("""
parameter: b = False
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=[]).run(mode="dryrun")
    assert env.sos_dict["b"] == False
    env.sos_dict.pop("b")
    Base_Executor(wf, args=["--b"]).run(mode="dryrun")
    assert env.sos_dict["b"] == True
    env.sos_dict.pop("b")
    Base_Executor(wf, args=["--no-b"]).run(mode="dryrun")
    assert env.sos_dict["b"] == False
    env.sos_dict.pop("b")
    #
    # parameters cannot coincide with a readonly global variable
    # are masked by previous definition
    script = SoS_Script("""
a = 4
parameter: a = 5
[0]
""")
    wf = script.workflow()
    with pytest.raises(Exception):
        Base_Executor(wf, args=["--a", 7])
    # assert env.sos_dict['a'], 4)
    #
    # test parameters with dash
    script = SoS_Script("""
parameter: a_b = 5
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--a_b", "10"]).run(mode="dryrun")
    assert env.sos_dict["a_b"] == 10
    env.sos_dict.pop("a_b")
    Base_Executor(wf, args=["--a-b", "10"]).run(mode="dryrun")
    assert env.sos_dict["a_b"] == 10
    env.sos_dict.pop("a_b")
    #
    #
    script = SoS_Script("""
parameter: a_b = int
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--a_b", "10"]).run(mode="dryrun")
    assert env.sos_dict["a_b"] == 10
    env.sos_dict.pop("a_b")
    Base_Executor(wf, args=["--a-b", "10"]).run(mode="dryrun")
    assert env.sos_dict["a_b"] == 10
    env.sos_dict.pop("a_b")
    #
    # test support for type path, paths, file_target and sos_targets
    script = SoS_Script("""
parameter: path_var = path
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
    assert isinstance(env.sos_dict["path_var"], path)
    #
    script = SoS_Script("""
parameter: path_var = file_target
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
    assert isinstance(env.sos_dict["path_var"], file_target)
    #
    script = SoS_Script("""
parameter: path_var = paths
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--path-var", "a.txt", "b.txt"]).run(mode="run")
    assert isinstance(env.sos_dict["path_var"], paths)
    #
    #
    script = SoS_Script("""
parameter: path_var = sos_targets
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
    assert isinstance(env.sos_dict["path_var"], sos_targets)

    #
    script = SoS_Script("""
parameter: path_var = path('a.txt')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
    assert isinstance(env.sos_dict["path_var"], path)
    #
    script = SoS_Script("""
parameter: path_var = file_target('a.txt')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
    assert isinstance(env.sos_dict["path_var"], file_target)
    #
    script = SoS_Script("""
parameter: path_var = paths('a.txt')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--path-var", "a.txt", "b.txt"]).run(mode="run")
    assert isinstance(env.sos_dict["path_var"], paths)
    #
    #
    script = SoS_Script("""
parameter: path_var = sos_targets('a.txt')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
    assert isinstance(env.sos_dict["path_var"], sos_targets)

    #
    # Test allow the use of sos keywords as parameters #1041
    script = SoS_Script("""\
[1]
parameter: input = 5
output = 10
python: expand=True
print({input})
""")
    wf = script.workflow()
    Base_Executor(wf).run()
    # multiple parameters
    script = SoS_Script("""
parameter: a_b = int
[0]
parameter: c_b = list
""")
    wf = script.workflow()
    assert sorted(list(wf.parameters().keys())) == ["a_b", "c_b"]


def test_param_in_task():
    """Test specification of parameters in tasks"""
    with pytest.raises(Exception):
        SoS_Script(
            """\
[1]
output: 'a.txt'
task:
parameter: value = 'a'
sh: expand=True
echo {value} >  a.txt
""",)
    # multiple parameters
    script = SoS_Script("""
parameter: a_b = int
[0]
parameter: c_b = list
""")
    wf = script.workflow()
    assert sorted(list(wf.parameters().keys())) == ["a_b", "c_b"]


def test_type_trait_parameter():
    # type trait
    script = SoS_Script("""
parameter: b
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
    assert env.sos_dict["b"] == "5"
    #
    script = SoS_Script("""
parameter: b :str
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
    assert env.sos_dict["b"] == "5"
    #
    script = SoS_Script("""
parameter: b : list
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
    assert env.sos_dict["b"] == ["5"]

    #
    script = SoS_Script("""
parameter: b : list = 5
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
    assert env.sos_dict["b"] == 5
    #
    script = SoS_Script("""
parameter: b : int
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
    assert env.sos_dict["b"] == 5


def test_input_target():
    # test input of targets
    script = SoS_Script("""
parameter: b : file_target
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["b"].__class__.__name__ == "file_target"
    #
    script = SoS_Script("""
parameter: b = file_target('file')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["b"].__class__.__name__ == "file_target"
    #
    script = SoS_Script("""
parameter: b : sos_targets
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "aaa", "bbb"]).run(mode="dryrun")
    assert env.sos_dict["b"].__class__.__name__ == "sos_targets"
    #
    script = SoS_Script("""
parameter: b = sos_targets('file')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["b"].__class__.__name__ == "sos_targets"
    #
    script = SoS_Script("""
parameter: a_b : file_target
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["a_b"].__class__.__name__ == "file_target"
    #
    script = SoS_Script("""
parameter: a_b = file_target('file')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["a_b"].__class__.__name__ == "file_target"
    #
    script = SoS_Script("""
parameter: a_b : sos_targets
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--a-b", "aaa", "bbb"]).run(mode="dryrun")
    assert env.sos_dict["a_b"].__class__.__name__ == "sos_targets"
    #
    script = SoS_Script("""
parameter: a_b = sos_targets('file')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["a_b"].__class__.__name__ == "sos_targets"
    #
    #
    #
    #
    script = SoS_Script("""
parameter: b : path
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["b"].__class__.__name__ == "path"
    #
    script = SoS_Script("""
parameter: b = path('file')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["b"].__class__.__name__ == "path"
    #
    script = SoS_Script("""
parameter: b : paths
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "aaa", "bbb"]).run(mode="dryrun")
    assert env.sos_dict["b"].__class__.__name__ == "paths"
    #
    script = SoS_Script("""
parameter: b = paths('file')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["b"].__class__.__name__ == "paths"
    #
    script = SoS_Script("""
parameter: a_b : path
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["a_b"].__class__.__name__ == "path"
    #
    script = SoS_Script("""
parameter: a_b = path('file')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["a_b"].__class__.__name__ == "path"
    #
    script = SoS_Script("""
parameter: a_b : paths
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--a-b", "aaa", "bbb"]).run(mode="dryrun")
    assert env.sos_dict["a_b"].__class__.__name__ == "paths"
    #
    script = SoS_Script("""
parameter: a_b = paths('file')
[0]
""")
    wf = script.workflow()
    Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
    assert env.sos_dict["a_b"].__class__.__name__ == "paths"
    #


def test_section_directives():
    """Test directives of sections"""
    # cannot be in the global section
    # multi-line OK
    SoS_Script("""
[0]
input: 'filename',
'filename1'

""")
    # An abusive case with multi-line OK, from first column ok
    SoS_Script("""
[0]
input: 'filename',
'filename1',
filename4,
opt1=value
output:
blah

depends:
'something else'
""")
    # option with expression ok
    SoS_Script("""
[0]
input: 'filename',  'filename2', opt=value==1

""")
    # unrecognized directive, allowed now
    SoS_Script("""
[0]
something: 'filename',  filename2, opt=value==1
""")
    # need commma
    with pytest.raises(ParsingError):
        SoS_Script("""
[0]
input: 'filename'  filename2
""",)
    # can be after action
    SoS_Script("""
[0]
func()
input: 'filename',  'filename2', opt=value==1
""")
    # assignments between directives are allowed
    SoS_Script("""
[0]
input: 'filename',  'filename2', opt=value==1
a = 'some text'
output: 'filename',  'filename2', opt=value==1
""")
    # can have action between directives
    SoS_Script("""
[0]
input: 'filename',  'filename2', opt=value==1
abc
output: 'filename',  'filename2', opt=value==1
""")


def test_script_format():
    """Test putting scripts directly under action"""
    script = SoS_Script("""
[0]
input: 'filename',  'filename2', opt=value==1
R:

open('something')
save.put()

""")
    script = SoS_Script("""
[0]
input: 'filename',  'filename2', opt=value==1
R: concurrent = True

open('something')
save.put()

""")
    script = SoS_Script("""
[0]
input: 'filename',  'filename2', opt=value==1
R: concurrent = True,
workdir = 'someelse else'

open('something')
save.put()

""")
    # test dedent
    script = SoS_Script("""
[0]
python3:
from sos.runtime import logger
logger.warning('I am from a dented text')
if 1:
    logger.warning('Another warning')
""")
    script.workflow()
    # with section head in the script,
    # this will not work even if the embedded
    # python script is perfectly valid.
    with pytest.raises(ParsingError):
        SoS_Script(
            '''
[0]
input: 'filename',  'filename2', opt=value==1
python3:

with open('something') as e:
e.write("""
[section]
""")

''',)
    # scripts with section head-like lines
    script = SoS_Script("""
[0]
R:
some::function(param)
""")
    script.workflow()
    #
    # script with first-line indent
    #
    script = SoS_Script("""
[0]
sh:
echo "a"

sh('echo "b"')
""")
    script.workflow()
    #
    # script with triple quote and format string
    # #1211
    script = SoS_Script('''
[1]
python3: expand = "${ }"

ld = ${'100'}
a =  """doc"""

''')
    wf = script.workflow()
    Base_Executor(wf).run()
    script = SoS_Script('''
[default]
report: expand = "${ }"
ld_file = ${_input['']:r}
""" \'\'\'
{}, ${_output:r},

''')
    wf = script.workflow()
    Base_Executor(wf).run()


def test_input(temp_factory):
    """Test input directive"""
    temp_factory("a.txt", "b.txt", "a.pdf", "a0", "a1")
    script = SoS_Script("""
[0]
files = ['a.txt', 'b.txt']

input: 'a.pdf', files

""")
    wf = script.workflow()
    Base_Executor(wf).run(mode="dryrun")
    #
    # test input types
    script = SoS_Script("""
[0:shared={'i':'_input', 'o':'_output'}]
files = (f"a{i}" for i in range(2))
input: {'a.txt', 'b.txt'}, files
output: (f"a{x}" for x in _input)

""")
    wf = script.workflow()
    Base_Executor(wf).run(mode="dryrun")
    assert sorted(env.sos_dict["i"]) == sos_targets(
        ["a.txt", "a0", "a1", "b.txt"])
    assert sorted(env.sos_dict["o"]) == sos_targets(
        ["aa.txt", "aa0", "aa1", "ab.txt"])


def test_group_by(temp_factory, clear_now_and_after):
    """Test group_by parameter of step input"""
    clear_now_and_after('a.txt', 'b.txt', 'c.txt' 'xx.txt')
    clear_now_and_after([f"b{x+1}.txt" for x in range(6)])
    # group_by = 'all'
    temp_factory([f"a{x}.txt" for x in range(15)])
    #
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: [f'a{x}.txt' for x in range(1, 5)], group_by='all'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("a1.txt", "a2.txt", "a3.txt", "a4.txt")
    ]
    assert env.sos_dict["executed"][0].labels == ["0"] * 4
    # group_by = 'single'
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='single'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("a1.txt"),
        sos_targets("a2.txt"),
        sos_targets("a3.txt"),
        sos_targets("a4.txt"),
    ]
    # group_by = 'pairs'
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: [f'a{x}.txt' for x in range(1, 5)], group_by='pairs'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("a1.txt", "a3.txt"),
        sos_targets("a2.txt", "a4.txt")
    ]
    # group_by = 'pairs2'
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: [f'a{x}.txt' for x in range(1, 9)], group_by='pairs2'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("a1.txt", "a2.txt", "a5.txt", "a6.txt"),
        sos_targets("a3.txt", "a4.txt", "a7.txt", "a8.txt"),
    ]
    # group_by = 'pairs3'
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: [f'a{x}.txt' for x in range(1, 13)], group_by='pairs3'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("a1.txt", "a2.txt", "a3.txt", "a7.txt", "a8.txt", "a9.txt"),
        sos_targets("a4.txt", "a5.txt", "a6.txt", "a10.txt", "a11.txt",
                    "a12.txt"),
    ]

    # group_by = 'pairwise'
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='pairwise'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("a1.txt", "a2.txt"),
        sos_targets("a2.txt", "a3.txt"),
        sos_targets("a3.txt", "a4.txt"),
    ]
    # group_by = 'pairwiseN'
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: ['a{}.txt'.format(x) for x in range(1, 7)], group_by='pairwise2'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("a1.txt", "a2.txt", "a3.txt", "a4.txt"),
        sos_targets("a3.txt", "a4.txt", "a5.txt", "a6.txt"),
    ], f'obtained {env.sos_dict["executed"]}'

    # group_by = 'combinations'
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='combinations'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"], [
        sos_targets("a1.txt", "a2.txt"),
        sos_targets("a1.txt", "a3.txt"),
        sos_targets("a1.txt", "a4.txt"),
        sos_targets("a2.txt", "a3.txt"),
        sos_targets("a2.txt", "a4.txt"),
        sos_targets("a3.txt", "a4.txt"),
    ]

    # group_by = 'combinations3'
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='combinations3'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets(["a1.txt", "a2.txt", "a3.txt"]),
        sos_targets(["a1.txt", "a2.txt", "a4.txt"]),
        sos_targets(["a1.txt", "a3.txt", "a4.txt"]),
        sos_targets(["a2.txt", "a3.txt", "a4.txt"]),
    ], f'obtained {env.sos_dict["executed"]}'
    # group_by chunks specified as integers
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: ['a{}.txt'.format(x) for x in range(1, 10)], group_by=3

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("a1.txt", "a2.txt", "a3.txt"),
        sos_targets("a4.txt", "a5.txt", "a6.txt"),
        sos_targets("a7.txt", "a8.txt", "a9.txt"),
    ]
    # group_by chunks specified as integer strings
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: ['a{}.txt'.format(x) for x in range(1, 10)], group_by='3'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("a1.txt", "a2.txt", "a3.txt"),
        sos_targets("a4.txt", "a5.txt", "a6.txt"),
        sos_targets("a7.txt", "a8.txt", "a9.txt"),
    ]
    # number of files should be divisible by group_by
    temp_factory(["a{}.txt".format(x) for x in range(1, 10)])
    execute_workflow(
        """
        [0]

        executed = []
        input: ['a{}.txt'.format(x) for x in range(1, 10)], group_by=4

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    # incorrect value causes an exception
    with pytest.raises(Exception):
        execute_workflow(
            """
            [0]

            executed = []
            input: ['a{}.txt'.format(x) for x in range(1, 10)], group_by='something'

            executed.append(_input)

            """,
            options={'run_mode': "dryrun"})

    #
    # group by label
    file_target("c.txt").touch()
    execute_workflow(
        """
        [A]
        output: 'a.txt'
        _output.touch()

        [B]
        input: for_each={'i': range(2)}
        output: 'b.txt', 'b1.txt', group_by=1
        _output.touch()

        [0: shared='executed']
        executed = []

        input: 'c.txt', output_from(['A', 'B']), group_by='label'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("c.txt"),
        sos_targets("a.txt"),
        sos_targets("b.txt", "b1.txt"),
    ]
    #
    # group_by='pairlabel'
    file_target("c.txt").touch()
    execute_workflow(
        """
        [A]
        output: 'a1.txt', 'a2.txt'
        _output.touch()

        [B]
        output: 'b1.txt', 'b2.txt'
        _output.touch()

        [0: shared='executed']
        executed = []

        input: output_from(['A', 'B']), group_by='pairlabel'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("a1.txt", "b1.txt"),
        sos_targets("a2.txt", "b2.txt")
    ]

    # group_by='pairlabel3'
    temp_factory(["c{}.txt".format(x) for x in range(1, 7)])

    execute_workflow(
        """
        [A]
        output: [f'a{x}.txt' for x in range(1, 7)]
        _output.touch()

        [B]
        output: [f'b{x}.txt' for x in range(1, 7)]
        _output.touch()

        [0: shared='executed']
        executed = []

        input: [f'c{x}.txt' for x in range(1, 7)], output_from(['A', 'B']), group_by='pairlabel3'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets(
            "c1.txt",
            "c2.txt",
            "c3.txt",
            "a1.txt",
            "a2.txt",
            "a3.txt",
            "b1.txt",
            "b2.txt",
            "b3.txt",
        ),
        sos_targets(
            "c4.txt",
            "c5.txt",
            "c6.txt",
            "a4.txt",
            "a5.txt",
            "a6.txt",
            "b4.txt",
            "b5.txt",
            "b6.txt",
        ),
    ]
    # group_by='pairlabel3'
    temp_factory(["c{}.txt".format(x) for x in range(1, 7)])

    execute_workflow(
        """
        [A]
        output: 'a1.txt'
        _output.touch()

        [B]
        output: [f'b{x}.txt' for x in range(1, 7)]
        _output.touch()

        [0: shared='executed']
        executed = []

        input: [f'c{x}.txt' for x in range(1, 3)], output_from(['A', 'B']), group_by='pairlabel3'

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("c1.txt", "a1.txt", "b1.txt", "b2.txt", "b3.txt"),
        sos_targets("c2.txt", "a1.txt", "b4.txt", "b5.txt", "b6.txt"),
    ]
    #
    # group by function
    file_target("c.txt").touch()
    execute_workflow(
        """
        [0: shared='executed']
        executed = []

        def grp(x):
            return  [x[:3], x[3:]]

        input: ['a{}.txt'.format(x) for x in range(5)], group_by=grp

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [["a0.txt", "a1.txt", "a2.txt"],
                                        ["a3.txt", "a4.txt"]]

    #
    # group by lambda function
    file_target("c.txt").touch()
    execute_workflow(
        """
        [0: shared='executed']
        executed = []

        input: ['a{}.txt'.format(x) for x in range(6)], group_by=lambda x: zip(x[:3], x[3:])

        executed.append(_input)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [["a0.txt", "a3.txt"],
                                        ["a1.txt", "a4.txt"],
                                        ["a2.txt", "a5.txt"]]


def test_output_group_by(temp_factory):
    """Test group_by parameter of step output"""
    # group_by = 'all'
    temp_factory(["a{}.txt".format(x) for x in range(4)])
    #
    execute_workflow(
        """
        [0: shared='executed']

        executed = []
        input: ['a{}.txt'.format(x) for x in range(4)], group_by=2
        output: ['a{}.txt.bak'.format(x) for x in range(4)], group_by=2

        executed.append(_output)

        """,
        options={'run_mode': "dryrun"})
    assert env.sos_dict["executed"] == [
        sos_targets("a0.txt.bak", "a1.txt.bak"),
        sos_targets("a2.txt.bak", "a3.txt.bak"),
    ]


def test_steps_with_step_name(clear_now_and_after):
    """Test from steps"""
    clear_now_and_after('a.txt')
    execute_workflow("""
        [step_10]

        output: 'a.txt'
        _output.touch()

        [step_20]
        input: output_from(step_name.split('_')[0] + '_10')
        print(_input)
        """)
    #
    execute_workflow("""
        [step_10]

        output: 'a.txt'
        _output.touch()

        [step_20]
        input: output_from(10)
        print(_input)
        """)


def test_section_actions():
    """Test actions of sections"""
    SoS_Script("""
[0]
func('''
multiline
string''', with_option=1
)
""")
    with pytest.raises(ParsingError):
        SoS_Script("""
[0]
func(
""",)


def test_longer_code():
    """Test definition of classes (with intermediate newlines) in step."""
    execute_workflow("""
        # first block
        [0: shared='b']
        class A:
            def __init__(self):
                pass

            # the newline above should be fine because SoS treat this as
            # regular lines
            def __call__(self):
                return 0

        b = A()()

        """)
    assert env.sos_dict["b"] == 0


def test_combined_workflow():
    """Test the creation and execution of combined workfow"""
    script = """
        a0 = 0
        if 'executed' not in locals():
            executed = []
        parameter: a = a0 + 1
        [a_1: shared='executed']
        executed.append(step_name)
        [a_2: shared='executed']
        executed.append(step_name)
        [a_3: shared='executed']
        executed.append(step_name)
        [a_4: shared='executed']
        executed.append(step_name)
        output: 'out_a_4'
        [b_1: shared=['executed', 'input_b1']]
        executed.append(step_name)
        input_b1 = _input
        [b_2: shared='executed']
        executed.append(step_name)
        [b_3: shared='executed']
        executed.append(step_name)
        [b_4: shared='executed']
        executed.append(step_name)
        [c: shared='executed']
        executed.append(step_name)
        [d: shared='executed']
        executed.append(step_name)
        """
    execute_workflow(script, workflow="a+b", options={'run_mode': 'dryrun'})
    assert env.sos_dict["executed"] == [
        "a_1", "a_2", "a_3", "a_4", "b_1", "b_2", "b_3", "b_4"
    ]

    assert env.sos_dict["a"] == 1
    assert env.sos_dict["input_b1"] == ["out_a_4"]
    #
    env.sos_dict.pop("executed", None)
    execute_workflow(
        script, workflow="a: 1-2 + a:4 + b:3-", options={'run_mode': 'dryrun'})
    assert env.sos_dict["executed"] == ["a_1", "a_2", "a_4", "b_3", "b_4"]
    #
    env.sos_dict.pop("executed", None)
    execute_workflow(script, workflow="a+c+d", options={'run_mode': 'dryrun'})
    assert env.sos_dict["executed"] == ["a_1", "a_2", "a_3", "a_4", "c", "d"]


def test_yaml_config(clear_now_and_after):
    """Test config file in yaml format"""
    clear_now_and_after('myconfig.yml', 'config.sos')
    with open("myconfig.yml", "w") as config:
        config.write("""
# Lines beginning with # are skipped when the JSON is parsed, so we can
# put comments into our JSON configuration files
{
    StoreOwner : "John Doe",

    # List of items that we sell
    Fruits: [ "apple", "banana", "pear" ],
    Price: 1.05
}
""")
    with open("config.sos", "w") as sos:
        sos.write("""
[0]
print(CONFIG['StoreOwner'])
print(CONFIG.get('StoreOwner', 'something'))
print(CONFIG.get('StoreOwnerSpouse', 'someone else'))
#print(CONFIG.StoreOwner)
""")
    # run the command
    assert subprocess.call(
        "sos run config.sos -c myconfig.yml",
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True,
    ) == 0
    # now test the value
    script = SoS_Script(filename="config.sos")
    wf = script.workflow()
    Base_Executor(wf, config={"config_file": "myconfig.yml"}).run()
    assert env.sos_dict["CONFIG"]["Price"] == 1.05
    assert env.sos_dict["CONFIG"]["StoreOwner"] == "John Doe"
    assert env.sos_dict["CONFIG"]["Fruits"] == ["apple", "banana", "pear"]


def test_var_output():
    """Test early appearance of variable output"""
    execute_workflow(
        """
        [0]
        seq = range(3)
        input: for_each='seq'
        output: f"test{_seq}.txt"
        print(_output)
        """,
        options={'run_mode': "dryrun"})


def test_cell():
    """Test ignoring %cell"""
    SoS_Script("""
%cell 1
[step ]
a = 1
""")


def test_overwrite_keyword(clear_now_and_after):
    """Test overwrite sos keyword with user defined one."""
    clear_now_and_after("a.txt")
    #
    execute_workflow("""
def run(script):
    pass

[1]
run:
    touch a.txt
""")
    assert not os.path.isfile("a.txt")
    #
    execute_workflow("""
parameter: run = 5

[1]
run:
    touch a.txt
""")


def test_help_message(sample_workflow):
    """Test help message from ipynb notebook"""
    msg = subprocess.check_output(
        f"sos run {sample_workflow} -h", shell=True).decode()
    assert "this comment will be included but not shown in help" not in msg
    assert msg.count(
        "this comment will become the comment for parameter b") == 1
    assert msg.count(
        "this comment will become the comment for parameter d") == 1
    assert msg.count("--c 3 (as int)") == 1
    assert msg.count("this is a section comment, will be displayed") == 1
    assert "this is a test workflow" not in msg
    assert "this is a cell with another kernel" not in msg
    assert "this comment will not be included in exported workflow" not in msg


def test_help_on_multi_workflow(script_factory):
    """Test help message of sos file (#985)"""
    test_sos = script_factory("""\
        [workflow_a_10,workflow_b]
        [workflow_a_20]
        [default]
        """)
    msg = subprocess.check_output(
        f"sos run {test_sos} -h", shell=True).decode()
    assert "workflow_a" in msg
    assert "workflow_b" in msg
    # when introducing sections
    assert "workflow_a_10, workflow_b" in msg
    assert "default" in msg


def test_parameter_abbreviation(clear_now_and_after):
    """Test potential problem caused by parameter abbreviation #1053"""
    clear_now_and_after("0914.txt")
    execute_workflow(
        """
        [global]
        parameter: name = '0914'

        [1]
        parameter: n = 4
        output: f'{name}.txt'
        print(_output)
        _output.touch()
        """,
        args=["--n", "5"])
    assert os.path.isfile("0914.txt")


def test_named_input(temp_factory, clear_now_and_after):
    """Test named input"""
    clear_now_and_after('c.txt')
    temp_factory('a.txt', content='a.txt' + '\n')
    temp_factory('b.txt', content='b.txt' + '\n')
    execute_workflow("""
        [1]
        input: {'A': 'a.txt', 'B': 'b.txt'}, group_by='pairlabel'
        output: 'c.txt'
        assert _input['A'] == [file_target('a.txt')]
        assert _input['B'] == [file_target('b.txt')]
        with open(_output, 'w') as out:
            out.write(open(_input['A']).read())
            out.write(open(_input['B']).read())
        """)
    assert open("c.txt").read() == "a.txt\nb.txt\n"


def test_named_output_in_depends(clear_now_and_after):
    """Test named_output in depends statement"""
    clear_now_and_after('a.txt')
    execute_workflow("""
        [A]
        output: A='a.txt'
        _output.touch()

        [default]
        depends: named_output('A')
        """)


def test_output_from_in_depends(clear_now_and_after):
    """Test output_from in depends statement"""
    clear_now_and_after('a.txt')
    execute_workflow("""
        [A]
        output: A='a.txt'
        _output.touch()

        [default]
        depends: output_from('A')
        """)


def test_named_output_in_output(clear_now_and_after):
    """Test named_output in output statement"""
    clear_now_and_after('a.txt')
    with pytest.raises(Exception):
        execute_workflow("""
            [A]
            output: Aa='a.txt'
            _output.touch()

            [default]
            output: named_output('Aa')
            """)


def test_output_from_in_output():
    """Test output_from in output statement"""
    with pytest.raises(Exception):
        execute_workflow("""
            [Aa]
            output: A='a.txt'
            _output.touch()

            [default]
            output: output_from('Aa')
            """)


def test_sos_step_in_input():
    """Test sos_step in input statement"""
    with pytest.raises(Exception):
        execute_workflow("""
            [A]
            output: A='a.txt'
            _output.touch()

            [default]
            input: sos_step('A')
            """)


def test_sos_variable_in_input():
    """Test sos_variable in input statement"""
    with pytest.raises(Exception):
        execute_workflow("""
            [A]
            a=1

            [default]
            input: sos_variable('a')
            """)


def test_sos_variable_in_output():
    """Test sos_variable in output statement"""

    with pytest.raises(Exception):
        execute_workflow("""
            [A]
            a = 1

            [default]
            output: sos_variable('a')
            """)


def test_sos_variable_with_keywordargument():
    """Test output_from in output statement"""

    with pytest.raises(Exception):
        execute_workflow("""
            [A]
            a = 1

            [default]
            depends: sos_variable(var='a')
            """)


def test_wide_card_step_name():
    """test resolving step name with *"""
    execute_workflow("""
        [A_1]

        [*_2]
        assert step_name == 'A_2', f'step_name is {step_name}, A_2 expected'
        """)


def test_outfrom_prev_step(clear_now_and_after):
    """Test output_from(-1) from output_from """
    clear_now_and_after('A_1.txt')
    execute_workflow("""
        [A_1]
        output: 'A_1.txt'
        _output.touch()

        [A_2]
        input: output_from(-1)

        [default]
        depends: sos_step('A_2')

        """)


def test_step_from_numeric_step():
    """Test sos_step(index) #1209"""
    execute_workflow("""
[1]

[2]
depends: sos_step('1')
""")

    #
    execute_workflow("""
[1]

[2]
depends: sos_step(1)

""")


def test_depends_on_step_with_unspecified_input(clear_now_and_after):
    clear_now_and_after("A_1.txt", "A_2.txt", "A_3.txt", "A_4.txt")
    #
    execute_workflow("""
        [A_1]
        output: f'{step_name}.txt'
        _output.touch()

        [A_2]
        output: f'{step_name}.txt'
        _output.touch()

        [A_3]
        output: f'{step_name}.txt'
        _output.touch()

        [A_4]
        output: f'{step_name}.txt'
        _output.touch()

        [default]
        depends: sos_step('A_2')
        """)
    assert os.path.isfile("A_1.txt")
    assert os.path.isfile("A_2.txt")
    assert not os.path.isfile("A_3.txt")
    assert not os.path.isfile("A_4.txt")
    #
    for file in ("A_1.txt", "A_2.txt", "A_3.txt", "A_4.txt"):
        if os.path.isfile(file):
            os.remove(file)
    #
    execute_workflow("""
        [A_1]
        output: f'{step_name}.txt'
        _output.touch()

        [A_2]
        output: f'{step_name}.txt'
        _output.touch()

        [A_3]
        output: f'{step_name}.txt'
        _output.touch()

        [A_4]
        output: f'{step_name}.txt'
        _output.touch()

        [default]
        input: output_from('A_2')
        """)
    assert os.path.isfile("A_1.txt")
    assert os.path.isfile("A_2.txt")
    assert not os.path.isfile("A_3.txt")
    assert not os.path.isfile("A_4.txt")
    #
    for file in ("A_1.txt", "A_2.txt", "A_3.txt", "A_4.txt"):
        if os.path.isfile(file):
            os.remove(file)
    #
    execute_workflow("""
        [A_1]
        output: f'{step_name}.txt'
        _output.touch()

        [A_2]
        input: None
        output: f'{step_name}.txt'
        _output.touch()

        [A_3]
        output: f'{step_name}.txt'
        _output.touch()

        [A_4]
        output: f'{step_name}.txt'
        _output.touch()

        [default]
        input: output_from('A_2')
        """)
    assert not os.path.isfile("A_1.txt")
    assert os.path.isfile("A_2.txt")
    assert not os.path.isfile("A_3.txt")
    assert not os.path.isfile("A_4.txt")


def test_output_from_workflow(clear_now_and_after):
    """Test output from workflow"""
    #
    #
    clear_now_and_after("A_1.txt", "A_2.txt")
    execute_workflow(r"""
        [A_1]
        output: f'{step_name}.txt'

        with open(_output, 'w') as out:
            out.write(step_name)
            out.write('\n')

        [A_2]
        output: f'{step_name}.txt'

        with open(_output, 'w') as out:
            out.write(step_name)
            out.write('\n')

        [default]
        input: output_from('A')

        with open(_input) as content:
            assert content.read() == 'A_2\n'

        """)
    # this should execute A_2 and A_1
    assert os.path.isfile("A_1.txt")
    assert os.path.isfile("A_2.txt")


def test_execute_global_section():
    """Global section should be executed only once #1219"""
    execute_workflow(r"""
        [global]
        parameter: A=[1, 2]
        parameter: B=[]
        B.extend(A)

        [default]
        assert B == [1, 2]
        input: for_each=dict(i=range(2))
        assert B == [1, 2]
        """)


def test_para_from_nested_workflow():
    """Test passing arguments to nested workflow #1229"""
    execute_workflow(
        r"""
        [A]
        parameter: param = str
        print(param)

        [B]
        print(step_name)

        [default]
        sos_run('B+A')
        """,
        args=["--param", "2"])


def test_indented_action():
    """Test the use of indented action in script format"""
    #
    # not indented
    execute_workflow(r"""
        [1]
        if True:
            sh:
        echo something

        [2]
        for i in range(2):
            sh:
        echo something
        """)
    # not indented double action
    with pytest.raises(Exception):
        execute_workflow(r"""
            [1]
            if True:
                sh:
            echo something

                python:
            print(1)
        """)
    # not indented Python structure
    with pytest.raises(Exception):
        execute_workflow(r"""
            [1]
            if True:
                sh:
            echo something
            else:
                python:
            print(1)
            """)
    #  indented
    execute_workflow(r"""
[1]
if True:
    sh:
        echo something

[2]
for i in range(2):
    sh:
        echo something
""")

    # indented double action
    execute_workflow(r"""
[1]
if True:
    sh:
        echo something
    python:
        print(1)
""")
    # indented Python structure
    execute_workflow(r"""
[1]
if True:
    sh:
        echo something
else:
    python:
        print(1)
""")
    # indented, nested structure
    execute_workflow(r"""
[1]
for i in range(2):
    if True:
        sh:
            echo something
    else:
        python:
            print(1)
""")
    #
    # wrong nested action
    execute_workflow(r"""
report:
    name: 'a.txt'
""")


def test_task_param_var():
    """Test global parameter passed to task parameters #1281"""
    execute_workflow(
        r"""
        [global]
        parameter: job_size = 60

        [1]
        task: trunk_size = job_size
        bash:
                echo 1
        """,
        options={"default_queue": "localhost"},
    )


def test_task_param_var_to_substep():
    """Test global parameter passed to task parameters in substep #1281"""
    execute_workflow(
        r"""
        [global]
        parameter: job_size = 60

        [1]
        input: for_each=dict(i=range(3))

        task: trunk_size = job_size
        bash: expand=True
                echo {i}
        """,
        options={"default_queue": "localhost"},
    )


def test_empty_parameter():
    # parameter: without content #1283
    execute_workflow(r"""
        parameter:

        print(1)
        """)


def test_sequential_substeps():
    """Test sequential execution of substeps"""
    execute_workflow(r"""
        [10: shared='sum']
        sum = 0
        input: for_each=dict(i=range(4)), concurrent=False
        sum += i
        print(f'sum is {sum} at index {_index}')
        """)
    assert env.sos_dict["sum"] == 6


def test_limited_concurrency():
    """Set concurrent=INT"""
    execute_workflow(r"""
        [10]
        input: for_each=dict(i=range(6)), concurrent=2
        print(i)
        """)


def test_concurrent_substep_with_step_import():
    """ Test concurrent substep with step leval import statement #1354"""
    execute_workflow("""
        import time
        input: for_each=dict(i=range(2))
        time.sleep(0)
        """)


def test_type_hint():
    """We should be able to differentiate type hint and sos action"""
    SoS_Script("""a : list = 5""")
    SoS_Script("""a : list""")
    # action
    SoS_Script("""a : input='filename' """)
    # action
    SoS_Script("""a : expand='${ }' """)


def test_sos_step_in_output():
    """Test sos_step in output statement"""
    with pytest.raises(Exception):
        execute_workflow("""
            [A]
            output: A='a.txt'
            _output.touch()

            [default]
            output: sos_step('A')
        """)


def test_comments(sample_workflow):
    """Test the use of comments in sos script"""
    # extract workflow from ipynb
    wf = extract_workflow(sample_workflow)
    assert "this is a test workflow" not in wf
    assert wf.count("this comment will be included but not shown in help") == 1
    assert wf.count("this comment will become the comment for parameter b") == 1
    assert wf.count("this comment will become the comment for parameter d") == 1
    assert "this is a cell with another kernel" not in wf
    assert "this comment will not be included in exported workflow" not in wf



import os
import subprocess

import pytest
from sos import execute_workflow
from sos._version import __version__
from sos.parser import SoS_Script
from sos.utils import env
# if the test is imported under sos/test, test interacive executor
from sos.workflow_executor import Base_Executor


def test_for_each_nested_list(temp_factory):
    """Test for_each option of input"""
    # test dictionary format of for_each
    temp_factory("p1.txt", "p2.txt", "p3.txt", "a.txt", "b.txt")
    # test same-level for loop and parameter with nested list
    execute_workflow(
        r"""
        [0: shared=['processed']]
        files = ['a.txt', 'b.txt']

        input: files, for_each={'par':[(1, 2), (1, 3), (2, 3)], 'res': ['p1.txt', 'p2.txt', 'p3.txt']}
        output: res

        processed = (par, res)
        """,
        options={"run_mode": "dryrun"},
    )
    assert env.sos_dict["step_processed"] == [
        ((1, 2), "p1.txt"),
        ((1, 3), "p2.txt"),
        ((2, 3), "p3.txt"),
    ]


def test_removal_of_output_from_failed_step(clear_now_and_after):
    """Test the removal of output files if a step fails #1055"""
    clear_now_and_after("failed.csv", "result.csv")
    script = """
[sub: provides='{file}.csv']
sh: expand=True
  touch {_output}
  eco "something wrong"

[step]
depends: 'failed.csv'
path('result.csv').touch()
"""
    with pytest.raises(Exception):
        execute_workflow(script, workflow='step')
    # rerun should still raise
    with pytest.raises(Exception):
        execute_workflow(script, workflow='step')

    assert not os.path.isfile("failed.csv")
    assert not os.path.isfile("result.csv")


def test_error_handling_of_substeps(clear_now_and_after):
    clear_now_and_after(
        [f"test_{i}.txt" for i in range(10)],
        [f"test_{i}.bak" for i in range(10)],
        "test_30.txt",
    )

    script = r"""
import time

[10]
input: for_each=dict(i=range(10)), concurrent=False
output: f'test_{i}.txt'

_output.touch()

fail_if(_index == 5, 'fail at 5')

[20]
output: _input.with_suffix('.bak')
_output.touch()

[30]
input: None
output: 'test_30.txt'
time.sleep(2)
_output.touch()
"""

    #
    # default mode
    #
    with pytest.raises(Exception):
        execute_workflow(script)

    for i in range(10):
        if i == 5:
            assert not os.path.isfile(f"test_{i}.txt")
            assert not os.path.isfile(f"test_{i}.bak")
        else:
            assert os.path.isfile(f"test_{i}.txt")
            # the following step is not executed
            assert not os.path.isfile(f"test_{i}.bak")
    # but the another branch continues
    assert os.path.isfile("test_30.txt")
    #
    # ignore mode
    #
    clear_now_and_after(
        [f"test_{i}.txt" for i in range(10)],
        [f"test_{i}.bak" for i in range(10)],
        "test_30.txt",
    )

    execute_workflow(script, options={"error_mode": "ignore"})
    for i in range(6, 10):
        if i == 5:
            assert not os.path.isfile(f"test_{i}.txt")
            assert not os.path.isfile(f"test_{i}.bak")
        else:
            assert os.path.isfile(f"test_{i}.txt")
            assert os.path.isfile(f"test_{i}.bak")
    assert os.path.isfile("test_30.txt")
    #
    # abort mode
    #
    clear_now_and_after(
        [f"test_{i}.txt" for i in range(10)],
        [f"test_{i}.bak" for i in range(10)],
        "test_30.txt",
    )

    with pytest.raises(Exception):
        execute_workflow(script, options={"error_mode": "abort"})
    for i in range(10):
        if i < 5:
            assert os.path.isfile(f"test_{i}.txt")
            assert not os.path.isfile(f"test_{i}.bak")
        else:
            assert not os.path.isfile(f"test_{i}.txt")
            assert not os.path.isfile(f"test_{i}.bak")
    assert not os.path.isfile("test_30.txt")


def test_parallel_nestedworkflow(clear_now_and_after):
    # 1375
    clear_now_and_after([f'{i+1}.txt' for i in range(5)])
    clear_now_and_after([f'{i+1}.out' for i in range(5)])
    execute_workflow(r"""
        [global]
        parameter: num = [x+1 for x in range(5)]

        [1]
        input: for_each = 'num'
        output: f'{_num}.txt'
        bash: expand = True
            touch {_output}

        [2]
        parameter: a = 1
        output: f'{_input:n}.out'
        sos_run('a' if a >=1 else 'b')

        [a,b]
        output: f'{_input:n}.out'
        bash: expand = True
            touch {_output}
        """)


def test_for_each_as_target_property_nested_list(temp_factory):
    """Test for_each option of input"""
    # test same-level for loop and parameter with nested list
    temp_factory("a.txt", "b.txt", "p1.txt", "p2.txt", "p3.txt")
    execute_workflow(
        r"""
        [0: shared=['processed']]
        files = ['a.txt', 'b.txt']
        processed = None

        input: files, for_each={'par':[(1, 2), (1, 3), (2, 3)], 'res': ['p1.txt', 'p2.txt', 'p3.txt']}
        output: _input.res

        processed = (_input.par, _input.res)
        """,
        options={"run_mode": "dryrun"},
    )

    assert env.sos_dict["step_processed"] == [
        ((1, 2), "p1.txt"),
        ((1, 3), "p2.txt"),
        ((2, 3), "p3.txt"),
    ]


def test_rerun_with_zap(clear_now_and_after):
    clear_now_and_after([f"zapped_example_{i}.txt.zapped" for i in range(3)])
    clear_now_and_after([f"zapped_example_{i}.bak" for i in range(3)])

    script = '''
        [step_10]
        input: for_each={'i': range(3)}
        output: f'zapped_example_{i}.txt'
        sh: expand=True
            echo "hello" > {_output}

        [step_20]
        input: group_by=1
        output: _input.with_suffix('.bak')
        sh: expand=True
            cp {_input} {_output}

        _input.zap()
    '''
    execute_workflow(script)
    execute_workflow(script)


def test_for_each_as_target_property_same_level_loop(temp_factory):
    """Test for_each option of input"""
    # test same-level for loop and parameter with nested list
    temp_factory("a.txt", "b.txt", "p1.txt", "p2.txt", "p3.txt")
    execute_workflow(
        r"""
        [0: shared=['processed']]
        files = ['a.txt', 'b.txt']
        par = [(1, 2), (1, 3), (2, 3)]
        res = ['p1.txt', 'p2.txt', 'p3.txt']

        input: files, for_each='par,res'
        output: res, group_by=1

        print([x._dict for x in step_input._groups])
        processed = (_input._par, _input._res)
        """,
        options={"run_mode": "dryrun"},
    )
    assert env.sos_dict["step_processed"] == [
        ((1, 2), "p1.txt"),
        ((1, 3), "p2.txt"),
        ((2, 3), "p3.txt"),
    ]



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import sys

__all__ = ["__version__", "SOS_FULL_VERSION"]

_py_ver = sys.version_info
if _py_ver.major == 2 or (_py_ver.major == 3 and
                          (_py_ver.minor, _py_ver.micro) < (6, 0)):
    raise SystemError(
        f"SOS requires Python 3.6 or higher. Please upgrade your Python {_py_ver.major}.{_py_ver.minor}.{_py_ver.micro}")

# version of the SoS language
__sos_version__ = "1.0"
# version of the sos command
__version__ = "0.23.2"
__py_version__ = f"{_py_ver.major}.{_py_ver.minor}.{_py_ver.micro}"

#
SOS_FULL_VERSION = f"{__version__} for Python {_py_ver.major}.{_py_ver.minor}.{_py_ver.minor}"



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import copy
import os
import pickle
import signal
import subprocess
import time
from collections import OrderedDict
from collections.abc import Mapping, Sequence
from contextlib import redirect_stderr, redirect_stdout
from threading import Event

import zmq

from .controller import (Controller, close_socket, connect_controllers,
                         create_socket, disconnect_controllers,
                         request_answer_from_controller,
                         send_message_to_controller)
from .eval import SoS_eval, SoS_exec
from .executor_utils import (__null_func__, clear_output, get_traceback_msg,
                             prepare_env)
from .messages import decode_msg
from .monitor import TaskMonitor
from .step_executor import parse_shared_vars
from .targets import (InMemorySignature, dynamic, file_target, path, sos_step,
                      sos_targets)
from .tasks import (TaskFile, combine_results, monitor_interval,
                    remove_task_files, resource_monitor_interval)
from .utils import (ProcessKilled, StopInputGroup, env, get_localhost_ip,
                    pickleable)


def signal_handler(*args, **kwargs):
    raise ProcessKilled()


class BaseTaskExecutor(object):
    """Task executor used to execute specified tasks. Any customized task executor
    should derive from this class.
    """

    def __init__(self):
        pass

    def execute(self, task_id):
        """Execute single or master task, return a dictionary"""
        tf = TaskFile(task_id)

        # this will automatically create a pulse file
        tf.status = "running"
        # write result file
        try:
            signal.signal(signal.SIGTERM, signal_handler)
            # open task file and get parameters
            params, runtime = tf.get_params_and_runtime()
            sig_content = tf.signature

            if env.config["verbosity"] is not None:
                env.verbosity = env.config["verbosity"]

            if env.config["run_mode"] == "dryrun":
                env.config["sig_mode"] = "ignore"

            env.logger.info(f"{task_id} ``started``")

            # the runtime consists of two parts, one is defined by the task (e.g.
            # parameter walltime) and saved in params.sos_dict['_runtime'], another
            # is defined by the task engine (e.g. max_walltime) saved in run_time.
            # max_walltime, max_mem, max_procs can only be defined in task_engine

            # let us make sure _runtime exists to avoid repeated check
            if "_runtime" not in params.sos_dict:
                params.sos_dict["_runtime"] = {}
            if "_runtime" not in runtime:
                runtime["_runtime"] = {}

            m = TaskMonitor(
                task_id,
                monitor_interval=monitor_interval,
                resource_monitor_interval=resource_monitor_interval,
                max_walltime=(runtime["_runtime"].get("max_walltime", None),
                              runtime["_runtime"].get("walltime", None)),
                max_mem=(runtime["_runtime"].get("max_mem", None),
                         runtime["_runtime"].get("mem", None)),
                max_procs=(runtime["_runtime"].get("max_procs", None),
                           runtime["_runtime"].get("procs", None)),
                sos_dict=params.sos_dict,
            )

            m.start()

            if hasattr(params, "task_stack"):
                res = self.execute_master_task(task_id, params, runtime,
                                               sig_content)
            else:
                res = self.execute_single_task(task_id, params, runtime,
                                               sig_content)
        except KeyboardInterrupt:
            tf.status = "aborted"
            raise
        except ProcessKilled as e:
            tf.status = "aborted"
            raise ProcessKilled("task interrupted") from e
        finally:
            signal.signal(signal.SIGTERM, signal.SIG_DFL)

        if res["ret_code"] != 0 and "exception" in res:
            with open(
                    os.path.join(
                        os.path.expanduser("~"), ".sos", "tasks",
                        task_id + ".soserr"),
                    "a",
            ) as err:
                err.write(f'Task {task_id} exits with code {res["ret_code"]}')

        if res.get("skipped", False):
            # a special mode for skipped to set running time to zero
            tf.status = "skipped"
        else:
            tf.add_outputs()
            tf.add_result(res)

            # this will remove pulse and other files
            tf.status = "completed" if res["ret_code"] == 0 else "failed"

        return res["ret_code"]

    def execute_single_task(self,
                            task_id,
                            params,
                            runtime,
                            sig_content,
                            quiet=False,
                            **kwargs):
        """
        Execute a single task, with

        task_id: id of the task, a string
        params: task definitions, with global_def, task, and sos_dict as its members.
        runtime: can have key '_runtime' for task specific runtime, and task_id for task specific variables
        sig_content: existing signatures

        All other task executors calls this function eventually.
        """
        # this is task specific runtime information, used to update global _runtime
        params.sos_dict["_runtime"].update(runtime["_runtime"])
        # this is subtask dictionary
        if task_id in runtime:
            params.sos_dict.update(runtime[task_id])

        if (quiet and "TASK" in env.config["SOS_DEBUG"] or
                "ALL" in env.config["SOS_DEBUG"]):
            env.log_to_file("TASK", f"Executing task {task_id}")

        global_def, task, sos_dict = params.global_def, params.task, params.sos_dict

        prepare_env(global_def[0], global_def[1])

        env.sos_dict.quick_update(sos_dict)

        for key in [
                "step_input",
                "_input",
                "step_output",
                "_output",
                "step_depends",
                "_depends",
        ]:
            if key in sos_dict and isinstance(sos_dict[key], sos_targets):
                # resolve remote() target
                env.sos_dict.set(
                    key, sos_dict[key].remove_targets(
                        type=sos_step).resolve_remote())

        # when no output is specified, we just treat the task as having no output (determined)
        env.sos_dict["_output"]._undetermined = False
        sig = (None
               if env.config["sig_mode"] == "ignore" else InMemorySignature(
                   env.sos_dict["_input"],
                   env.sos_dict["_output"],
                   env.sos_dict["_depends"],
                   env.sos_dict["__signature_vars__"],
                   shared_vars=parse_shared_vars(env.sos_dict["_runtime"].get(
                       "shared", None)),
               ))

        if sig and self._validate_task_signature(
                sig, sig_content.get(task_id, {}), task_id, quiet):
            # env.logger.info(f'{task_id} ``skipped``')
            return self._collect_task_result(
                task_id, sos_dict, params.tags, skipped=True, signature=sig)

        # if we are to really execute the task, touch the task file so that sos status shows correct
        # execution duration.
        if not quiet:
            sos_dict["start_time"] = time.time()

        # task output
        env.sos_dict.set(
            "__std_out__",
            os.path.join(
                os.path.expanduser("~"), ".sos", "tasks", task_id + ".sosout"),
        )
        env.sos_dict.set(
            "__std_err__",
            os.path.join(
                os.path.expanduser("~"), ".sos", "tasks", task_id + ".soserr"),
        )
        env.logfile = os.path.join(
            os.path.expanduser("~"), ".sos", "tasks", task_id + ".soserr")
        # clear the content of existing .out and .err file if exists, but do not create one if it does not exist
        if os.path.exists(env.sos_dict["__std_out__"]):
            open(env.sos_dict["__std_out__"], "w").close()
        if os.path.exists(env.sos_dict["__std_err__"]):
            open(env.sos_dict["__std_err__"], "w").close()

        orig_dir = os.getcwd()
        try:
            # go to 'workdir'
            if "workdir" in sos_dict["_runtime"]:
                workdir = path(sos_dict["_runtime"]["workdir"])
                if not os.path.isdir(workdir):
                    try:
                        os.makedirs(workdir)
                        os.chdir(workdir)
                    except Exception as e:
                        # sometimes it is not possible to go to a "workdir" because of
                        # file system differences, but this should be ok if a work_dir
                        # has been specified.
                        env.logger.debug(
                            f'Failed to create workdir {sos_dict["_runtime"]["workdir"]}: {e}'
                        )
                else:
                    os.chdir(workdir)

            # we will need to check existence of targets because the task might
            # be executed on a remote host where the targets are not available.
            for target in (sos_dict["_input"] if isinstance(
                    sos_dict["_input"], list) else
                           []) + (sos_dict["_depends"] if isinstance(
                               sos_dict["_depends"], list) else []):
                # if the file does not exist (although the signature exists)
                # request generation of files
                if isinstance(target, str):
                    if not file_target(target).target_exists("target"):
                        # remove the signature and regenerate the file
                        raise RuntimeError(f"{target} not found")
                # the sos_step target should not be checked in tasks because tasks are
                # independently executable units.
                elif not isinstance(
                        target,
                        sos_step) and not target.target_exists("target"):
                    raise RuntimeError(f"{target} not found")

            # create directory. This usually has been done at the step level but the task can be executed
            # on a remote host where the directory does not yet exist.
            ofiles = env.sos_dict["_output"]
            if ofiles.valid():
                for ofile in ofiles:
                    parent_dir = ofile.parent
                    if not parent_dir.is_dir():
                        parent_dir.mkdir(parents=True, exist_ok=True)

            # set environ ...
            # we join PATH because the task might be executed on a different machine
            if "env" in sos_dict["_runtime"]:
                for key, value in sos_dict["_runtime"]["env"].items():
                    if "PATH" in key and key in os.environ:
                        new_path = OrderedDict()
                        for p in value.split(os.pathsep):
                            new_path[p] = 1
                        for p in value.split(os.environ[key]):
                            new_path[p] = 1
                        os.environ[key] = os.pathsep.join(new_path.keys())
                    else:
                        os.environ[key] = value
            if "prepend_path" in sos_dict["_runtime"]:
                if isinstance(sos_dict["_runtime"]["prepend_path"], str):
                    os.environ["PATH"] = (
                        sos_dict["_runtime"]["prepend_path"] + os.pathsep +
                        os.environ["PATH"])
                elif isinstance(env.sos_dict["_runtime"]["prepend_path"],
                                Sequence):
                    os.environ["PATH"] = (
                        os.pathsep.join(sos_dict["_runtime"]["prepend_path"]) +
                        os.pathsep + os.environ["PATH"])
                else:
                    raise ValueError(
                        f'Unacceptable input for option prepend_path: {sos_dict["_runtime"]["prepend_path"]}'
                    )

            with open(env.sos_dict["__std_out__"],
                      "a") as my_stdout, open(env.sos_dict["__std_err__"],
                                              "a") as my_stderr:
                with redirect_stdout(my_stdout), redirect_stderr(my_stderr):
                    # step process
                    SoS_exec(task)

                if "logfile" in params.sos_dict["_runtime"]:
                    logfile = params.sos_dict["_runtime"]["logfile"]
                    if not os.path.isfile(logfile):
                        raise ValueError(
                            f"logfile {logfile} does not exist after the completion of task"
                        )
                    try:
                        with open(logfile, "r") as log:
                            my_stdout.write(f"logfile: {logfile}\n")
                            my_stdout.write(log.read())
                    except Exception as e:
                        raise ValueError(
                            f"Failed to collect logfile {logfile} after the completion of task: {e}"
                        ) from e

            if quiet or env.config["run_mode"] != "run":
                env.logger.debug(f"{task_id} ``completed``")
            else:
                env.logger.info(f"{task_id} ``completed``")

        except StopInputGroup as e:
            # task ignored with stop_if exception
            if not e.keep_output:
                clear_output()
                env.sos_dict["_output"] = sos_targets([])
            if e.message:
                env.logger.info(e.message)
            return {
                "ret_code": 0,
                "task": task_id,
                "tags": params.tags,
                "input": sos_targets([]),
                "output": env.sos_dict["_output"],
                "depends": sos_targets([]),
                "shared": {},
            }
        except KeyboardInterrupt:
            env.logger.error(f"{task_id} ``interrupted``")
            raise
        except subprocess.CalledProcessError as e:
            return {
                "ret_code": e.returncode,
                "task": task_id,
                "tags": params.tags,
                "shared": {},
                "exception": RuntimeError(e.stderr),
            }
        except ProcessKilled:
            env.logger.error(f"{task_id} ``interrupted``")
            raise
        except Exception as e:
            msg = get_traceback_msg(e)
            # env.logger.error(f'{task_id} ``failed``: {msg}')
            with open(
                    os.path.join(
                        os.path.expanduser("~"), ".sos", "tasks",
                        task_id + ".soserr"),
                    "a",
            ) as err:
                err.write(msg + "\n")
            return {
                "ret_code": 1,
                "exception": RuntimeError(msg),
                "task": task_id,
                "tags": params.tags,
                "shared": {},
            }
        finally:
            os.chdir(orig_dir)

        return self._collect_task_result(
            task_id, sos_dict, params.tags, signature=sig)

    def execute_master_task(self, task_id, params, master_runtime, sig_content):
        """
        Execute a master task with multiple subtasks.

        task_id; id of master task.
        params: master parameters, with params.task_stack having params and
            runtime of subtasks
        master_runtime: master runtime, with runtime supplemented by subtasks
        sig_content: master signature with signatures for all subtasks.

        """
        # used by self._collect_subtask_outputs
        self.master_stdout = os.path.join(
            os.path.expanduser("~"), ".sos", "tasks", task_id + ".sosout")
        self.master_stderr = os.path.join(
            os.path.expanduser("~"), ".sos", "tasks", task_id + ".soserr")

        if os.path.exists(self.master_stdout):
            open(self.master_stdout, "w").close()
        if os.path.exists(self.master_stderr):
            open(self.master_stderr, "w").close()

        # options specified from command line, most likely from cluster system
        #
        # As explained in ticket 1304:
        #
        # From the point of cluster system, this is only one "job" on each
        # node. The "n_procs" is the "number of cpus per job per node", which,
        # is "cores" (per task) * "n_workers" (per node). Therefore,
        # "n_procs" should not be used.
        n_nodes, _ = self._parse_num_workers(env.config["worker_procs"])

        # regular trunk_workers = ?? (0 was used as default)
        # a previous version of master task file has params.num_workers
        n_workers = (
            params.num_workers if hasattr(params, "num_workers") else
            params.sos_dict["_runtime"].get("num_workers", 1))

        if isinstance(n_workers, int):
            if n_workers > 1:
                results = self.execute_master_task_in_parallel(
                    params, master_runtime, sig_content, n_workers)
            else:
                # n_workers = 1
                results = self.execute_master_task_sequentially(
                    params, master_runtime, sig_content)
        elif n_nodes == 1:
            if n_workers is None:
                n_workers = 1
            elif isinstance(n_workers, Sequence):
                # single node, so n_workers can have at most 1 element
                n_workers = n_workers[0]
            elif not isinstance(n_workers, int):
                raise ValueError(f"Illegal number of workers {n_workers}")

            if n_workers == 1:
                results = self.execute_master_task_sequentially(
                    params, master_runtime, sig_content)
            else:
                results = self.execute_master_task_in_parallel(
                    params, master_runtime, sig_content, n_workers)
        else:
            # n_workers should be a sequence
            if not n_workers:
                n_workers = env.config["worker_procs"]
            elif len(n_workers) != n_nodes:
                env.logger.warning(
                    f'task options trunk_workers={n_workers} is inconsistent with command line option -j {env.config["worker_procs"]}'
                )
            results = self.execute_master_task_distributedly(
                params, master_runtime, sig_content, n_workers)

        return combine_results(task_id, results)

    def execute_master_task_in_parallel(self, params, master_runtime,
                                        sig_content, n_workers):
        # multiple workers, concurrent execution using a pool
        from multiprocessing.pool import Pool

        p = Pool(n_workers)
        results = []
        for sub_id, sub_params in params.task_stack:
            if hasattr(params, "common_dict"):
                sub_params.sos_dict.update(params.common_dict)
            sub_runtime = {
                x: master_runtime.get(x, {}) for x in ("_runtime", sub_id)
            }
            sub_sig = {sub_id: sig_content.get(sub_id, {})}
            res = p.apply_async(
                self.execute_single_task,
                (sub_id, sub_params, sub_runtime, sub_sig, True),
                callback=self._append_subtask_outputs,
            )
            results.append(res)

        for idx, r in enumerate(results):
            results[idx] = r.get()
            self._cache_subresult(params.ID, results[idx])
        p.close()
        p.join()
        return results

    def execute_master_task_sequentially(self, params, master_runtime,
                                         sig_content):
        # single worker, execute sequentially, n_workers is not a positive number
        results = []
        for sub_id, sub_params in params.task_stack:
            if hasattr(params, "common_dict"):
                sub_params.sos_dict.update(params.common_dict)
            sub_runtime = {
                x: master_runtime.get(x, {}) for x in ("_runtime", sub_id)
            }
            sub_sig = {sub_id: sig_content.get(sub_id, {})}
            res = self.execute_single_task(sub_id, sub_params, sub_runtime,
                                           sub_sig, True)
            try:
                self._append_subtask_outputs(res)
            except Exception as e:
                env.logger.warning(
                    f"Failed to copy result of subtask {sub_id}: {e}")
            self._cache_subresult(params.ID, res)
            results.append(res)
        return results

    def execute_master_task_distributedly(self, params, master_runtime,
                                          sig_content, n_workers):
        # multiple workers, start workers from remote hosts
        env.zmq_context = zmq.Context()

        # control panel in a separate thread, connected by zmq socket
        ready = Event()
        self.controller = Controller(ready)
        self.controller.start()
        # wait for the thread to start with a signature_req saved to env.config
        ready.wait()

        connect_controllers(env.zmq_context)

        succ = True
        try:
            # start a result receving socket
            self.result_pull_socket = create_socket(env.zmq_context, zmq.PULL,
                                                    "substep result collector")
            local_ip = get_localhost_ip()
            port = self.result_pull_socket.bind_to_random_port(
                f"tcp://{local_ip}")
            env.config["sockets"][
                "result_push_socket"] = f"tcp://{local_ip}:{port}"

            # send tasks to the controller
            results = []
            for sub_id, sub_params in params.task_stack:
                if hasattr(params, "common_dict"):
                    sub_params.sos_dict.update(params.common_dict)
                sub_runtime = {
                    x: master_runtime.get(x, {}) for x in ("_runtime", sub_id)
                }
                sub_sig = {sub_id: sig_content.get(sub_id, {})}

                # submit tasks
                send_message_to_controller([
                    "task",
                    dict(
                        task_id=sub_id,
                        params=sub_params,
                        runtime=sub_runtime,
                        sig_content=sub_sig,
                        config=env.config,
                        quiet=True,
                    ),
                ])

            for _ in range(len(params.task_stack)):
                res = decode_msg(self.result_pull_socket.recv())
                try:
                    self._append_subtask_outputs(res)
                except Exception as e:
                    env.logger.warning(f"Failed to copy result of subtask: {e}")
                self._cache_subresult(params.ID, res)
                results.append(res)
        except Exception as e:
            env.logger.error(f"Failed to execute task {params.ID}: {e}")
            succ = False
        finally:
            # end progress bar when the master workflow stops
            close_socket(self.result_pull_socket)
            env.log_to_file("EXECUTOR", f"Stop controller from {os.getpid()}")
            request_answer_from_controller(["done", succ])
            env.log_to_file("EXECUTOR", "disconntecting master")
            self.controller.join()
            disconnect_controllers(env.zmq_context if succ else None)
        return results

    def _cache_subresult(self, master_id, sub_result):
        cache_file = os.path.join(
            os.path.expanduser("~"), ".sos", "tasks", master_id + ".cache")
        with open(cache_file, "ab") as cache:
            pickle.dump(sub_result, cache)

    def _parse_num_workers(self, num_workers):
        # return number of nodes and workers
        if isinstance(num_workers, Sequence):
            if len(num_workers) >= 1:
                val = num_workers[0]
                if ":" in val:
                    val = val.rsplit(":", 1)[-1]
                n_workers = int(val.rsplit(":", 1)[-1])
                return len(num_workers), None if n_workers <= 0 else n_workers
            return None, None
        if isinstance(num_workers, str):
            if ":" in num_workers:
                num_workers = num_workers.rsplit(":", 1)[-1]
            n_workers = int(num_workers.rsplit(":", 1)[-1])
            return 1, None if n_workers <= 0 else n_workers
        if isinstance(num_workers, int) and num_workers >= 1:
            return 1, num_workers
        if num_workers is None:
            return None, None
        raise RuntimeError(
            f"Unacceptable value for parameter trunk_workers {num_workers}")

    def _append_subtask_outputs(self, result):
        """
        Append result returned from subtask to stdout and stderr streams
        """
        tid = result["task"]
        tags = " ".join(result.get("tags", []))
        with open(self.master_stdout,
                  "ab") as out, open(self.master_stderr, "ab") as err:
            dest = out if result["ret_code"] == 0 else err
            dest.write(
                f'> {tid}\t{tags}\t{"completed" if result["ret_code"] == 0 else "failed"}\n'
                .encode())
            if "output" in result and result["output"]:
                dest.write(f'output files:\n{result["output"]}\n'.encode())
            sub_out = os.path.join(
                os.path.expanduser("~"), ".sos", "tasks", tid + ".sosout")
            if os.path.isfile(sub_out):
                with open(sub_out, "rb") as sout:
                    out_content = sout.read()
                    if out_content:
                        dest.write(b"stdout:\n")
                        dest.write(out_content)
                try:
                    os.remove(sub_out)
                except Exception as e:
                    env.logger.warning(f"Failed to remove {sub_out}: {e}")

            sub_err = os.path.join(
                os.path.expanduser("~"), ".sos", "tasks", tid + ".soserr")
            if "exception" in result:
                dest.write(str(result["exception"]).encode())
            # err.write(
            #     f'{tid}: {"completed" if result["ret_code"] == 0 else "failed"}\n'
            #     .encode())
            if os.path.isfile(sub_err):
                with open(sub_err, "rb") as serr:
                    err_content = serr.read()
                    if err_content:
                        dest.write(b"stderr:\n")
                        dest.write(err_content)
                try:
                    os.remove(sub_err)
                except Exception as e:
                    env.logger.warning(f"Failed to remove {sub_err}: {e}")
            dest.write(b"\n")
        # remove other files as well
        try:
            remove_task_files(tid, [".sosout", ".soserr", ".out", ".err"])
        except Exception as e:
            env.logger.debug(f"Failed to remove files {tid}: {e}")

    def _collect_task_result(self,
                             task_id,
                             sos_dict,
                             tags,
                             skipped=False,
                             signature=None):
        shared = {}
        if "shared" in env.sos_dict["_runtime"]:
            svars = env.sos_dict["_runtime"]["shared"]
            if isinstance(svars, str):
                if svars not in env.sos_dict:
                    raise ValueError(
                        f"Unavailable shared variable {svars} after the completion of task {task_id}"
                    )
                if not pickleable(env.sos_dict[svars], svars):
                    env.logger.warning(
                        f"{svars} of type {type(env.sos_dict[svars])} is not sharable"
                    )
                else:
                    shared[svars] = copy.deepcopy(env.sos_dict[svars])
            elif isinstance(svars, Mapping):
                for var, val in svars.items():
                    if var != val:
                        env.sos_dict.set(var, SoS_eval(val))
                    if var not in env.sos_dict:
                        raise ValueError(
                            f"Unavailable shared variable {var} after the completion of task {task_id}"
                        )
                    if not pickleable(env.sos_dict[var], var):
                        env.logger.warning(
                            f"{var} of type {type(env.sos_dict[var])} is not sharable"
                        )
                    else:
                        shared[var] = copy.deepcopy(env.sos_dict[var])
            elif isinstance(svars, Sequence):
                # if there are dictionaries in the sequence, e.g.
                # shared=['A', 'B', {'C':'D"}]
                for item in svars:
                    if isinstance(item, str):
                        if item not in env.sos_dict:
                            raise ValueError(
                                f"Unavailable shared variable {item} after the completion of task {task_id}"
                            )
                        if not pickleable(env.sos_dict[item], item):
                            env.logger.warning(
                                f"{item} of type {type(env.sos_dict[item])} is not sharable"
                            )
                        else:
                            shared[item] = copy.deepcopy(env.sos_dict[item])
                    elif isinstance(item, Mapping):
                        for var, val in item.items():
                            if var != val:
                                env.sos_dict.set(var, SoS_eval(val))
                            if var not in env.sos_dict:
                                raise ValueError(
                                    f"Unavailable shared variable {var} after the completion of task {task_id}"
                                )
                            if not pickleable(env.sos_dict[var], var):
                                env.logger.warning(
                                    f"{var} of type {type(env.sos_dict[var])} is not sharable"
                                )
                            else:
                                shared[var] = copy.deepcopy(env.sos_dict[var])
                    else:
                        raise ValueError(
                            f"Option shared should be a string, a mapping of expression, or a list of string or mappings. {svars} provided"
                        )
            else:
                raise ValueError(
                    f"Option shared should be a string, a mapping of expression, or a list of string or mappings. {svars} provided"
                )
            env.log_to_file(
                "TASK",
                f'task {task_id} (index={env.sos_dict["_index"]}) return shared variable {shared}',
            )

        # the difference between sos_dict and env.sos_dict is that sos_dict (the original version) can have remote() targets
        # which should not be reported.
        if env.sos_dict["_output"].undetermined():
            # re-process the output statement to determine output files
            args, _ = SoS_eval(
                f'__null_func__({env.sos_dict["_output"]._undetermined})',
                extra_dict={"__null_func__": __null_func__},
            )
            # handle dynamic args
            env.sos_dict.set(
                "_output",
                sos_targets([
                    x.resolve() if isinstance(x, dynamic) else x for x in args
                ]),
            )

        return {
            "ret_code": 0,
            "task": task_id,
            "tags": tags,
            "input": sos_dict["_input"],
            "output": sos_dict["_output"],
            "depends": sos_dict["_depends"],
            "shared": shared,
            "skipped": skipped,
            "start_time": sos_dict.get("start_time", ""),
            "peak_cpu": sos_dict.get("peak_cpu", 0),
            "peak_mem": sos_dict.get("peak_mem", 0),
            "end_time": time.time(),
            "signature": {
                task_id: signature.write()
            } if signature else {},
        }

    def _validate_task_signature(self, sig, saved_sig, task_id, is_subtask):
        idx = env.sos_dict["_index"]
        if env.config["sig_mode"] in ("default", "skip", "distributed"):
            matched = sig.validate(saved_sig)
            if isinstance(matched, dict):
                # in this case, an Undetermined output can get real output files
                # from a signature
                env.sos_dict.set("_input", sos_targets(matched["input"]))
                env.sos_dict.set("_depends", sos_targets(matched["depends"]))
                env.sos_dict.set("_output", sos_targets(matched["output"]))
                env.sos_dict.update(matched["vars"])
                if is_subtask or env.config["run_mode"] != "run":
                    env.logger.debug(
                        f'Task ``{task_id}`` for substep ``{env.sos_dict["step_name"]}`` (index={idx}) is ``ignored`` due to saved signature'
                    )
                else:
                    env.logger.info(
                        f'Task ``{task_id}`` for substep ``{env.sos_dict["step_name"]}`` (index={idx}) is ``ignored`` due to saved signature'
                    )
                return True
        elif env.config["sig_mode"] == "assert":
            matched = sig.validate(saved_sig)
            if isinstance(matched, str):
                raise RuntimeError(f"Signature mismatch: {matched}")
            env.sos_dict.set("_input", sos_targets(matched["input"]))
            env.sos_dict.set("_depends", sos_targets(matched["depends"]))
            env.sos_dict.set("_output", sos_targets(matched["output"]))
            env.sos_dict.update(matched["vars"])

            if is_subtask or env.config["run_mode"] != "run":
                env.logger.debug(
                    f'Task ``{task_id}`` for substep ``{env.sos_dict["step_name"]}`` (index={idx}) is ``ignored`` with matching signature'
                )
            else:
                env.logger.info(
                    f'Task ``{task_id}`` for substep ``{env.sos_dict["step_name"]}`` (index={idx}) is ``ignored`` with matching signature'
                )
            sig.content = saved_sig
            return True
        elif env.config["sig_mode"] == "build":
            # The signature will be write twice
            if sig.write():
                if is_subtask or env.config["run_mode"] != "run":
                    env.logger.debug(
                        f'Task ``{task_id}`` for substep ``{env.sos_dict["step_name"]}`` (index={idx}) is ``ignored`` with signature constructed'
                    )
                else:
                    env.logger.info(
                        f'Task ``{task_id}`` for substep ``{env.sos_dict["step_name"]}`` (index={idx}) is ``ignored`` with signature constructed'
                    )
                return True
            return False
        elif env.config["sig_mode"] == "force":
            return False
        else:
            raise RuntimeError(
                f'Unrecognized signature mode {env.config["sig_mode"]}')



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil

import pytest
from sos import execute_workflow


@pytest.mark.skipif(not shutil.which('ruby'), reason='ruby not installed')
def test_ruby(clear_now_and_after):
    '''Test action ruby'''
    clear_now_and_after('sample.txt')
    execute_workflow(r'''
        [10]

        ruby:
        fname = "sample.txt"
        somefile = File.open(fname, "w")
        somefile.puts "Hello file!"
        somefile.close
        ''')
    assert os.path.isfile('sample.txt')



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil
import sys

import pytest
from sos import execute_workflow


@pytest.mark.skipif(
    not shutil.which('singularity'),
    reason='Skip test because docker is not installed.')
def test_bash_in_singularity():
    '''Test action bash in singularity environment'''
    execute_workflow(r'''
        [0]
        #run:  container='shub://singularityhub/ubuntu'
        run:  container='docker://ubuntu', engine='singularity'
        echo 'Echo'
        ''')


@pytest.mark.skipif(
    not shutil.which('singularity') or sys.platform == 'win32' or
    'TRAVIS' in os.environ or 'APPVEYOR' in os.environ,
    reason='Skip test because docker is not installed.')
def test_singularity_build_linux_image(self):
    '''Test action singularity build, needs authentication so no travis test'''
    execute_workflow(r'''
        singularity_build: dest='lolcow.simg', sudo=True, notest=True
        Bootstrap: docker
        From: ubuntu:16.04
        %post
        apt-get -y update
        apt-get -y install fortune cowsay lolcat
        %environment
        export LC_ALL=C
        export PATH=/usr/games:$PATH
        %runscript
        fortune | cowsay | lolcat
        ''')


@pytest.mark.skipif(
    not shutil.which('singularity') or sys.platform == 'win32' or
    'TRAVIS' in os.environ or 'APPVEYOR' in os.environ,
    reason='Skip test because docker is not installed.')
def test_singularity_build_from_shub(self):
    # needs authentication and cannot test on travis
    execute_workflow(r'''
        singularity_build(src='shub://GodloveD/lolcow', dest='lolcow_shub.simg', sudo=True, notest=True, force=True)
        ''')


@pytest.mark.skipif(
    not shutil.which('singularity') or sys.platform == 'win32',
    reason='Skip test because docker is not installed.')
def test_singularity_build_from_docker(self):
    execute_workflow(r'''
        singularity_build(src='docker://godlovedc/lolcow', dest='lolcow_docker.simg', sudo=True, notest=True, force=True)
        ''')
    # test handling simg with singularity
    execute_workflow(r'''
        run: container='lolcow_docker.simg'
        ls /
        ''')



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import pytest
from sos import execute_workflow
from sos.targets import file_target


def test_py_module(clear_now_and_after):
    '''Test target Py_Module'''
    clear_now_and_after('report.md')
    execute_workflow(r'''
[10]
depends: Py_Module('tabulate', autoinstall=True)
from tabulate import tabulate
table = [["Sun",696000,1989100000],["Earth",6371,5973.6],
    ["Moon",1737,73.5],["Mars",3390,641.85]]
report: output='report.md', expand=True
    {tabulate(table)}
''')
    assert file_target('report.md').target_exists('target')
    with open('report.md') as rep:
        assert rep.read().strip() == '''
-----  ------  -------------
Sun    696000     1.9891e+09
Earth    6371  5973.6
Moon     1737    73.5
Mars     3390   641.85
-----  ------  -------------
'''.strip()


def test_py_module_with_version():
    '''Test target Py_Module'''
    with pytest.raises(Exception):
        execute_workflow(r'''
[10]
depends: Py_Module('tabulate', version='2.0', autoinstall=True)
''')

        #
    with pytest.raises(Exception):
        execute_workflow(r'''
[10]
depends: Py_Module('tabulate>=2.0')
''')
        #

    with pytest.raises(Exception):
        execute_workflow(r'''
[10]
depends: Py_Module('tabulate==20.0')
''')
        #
    execute_workflow(r'''
[10]
depends: Py_Module('tabulate<2.0')
''')


def test_upgrade_py_module():
    '''Test upgrade py module #1246'''
    # first install tabulate == 0.7.5
    execute_workflow(r'''
[10]
depends: Py_Module('tabulate==0.7.5', autoinstall=True)
''')
    # test should pass
    execute_workflow(r'''
[10]
depends: Py_Module('tabulate'), Py_Module('tabulate==0.7.5')
''')
    # test for newer version should fail

    with pytest.raises(Exception):
        execute_workflow(r'''
[10]
depends: Py_Module('tabulate==0.8.3')
''')

    # auto install should work
    execute_workflow(r'''
[10]
depends: Py_Module('tabulate==0.8.3', autoinstall=True)
''')

    # test for old version should fail
    execute_workflow(r'''
        [10]
        depends: Py_Module('tabulate'), Py_Module('tabulate==0.8.3')
        ''')
    # test for old version should fail
    with pytest.raises(Exception):
        execute_workflow(r'''
        [10]
        depends: Py_Module('tabulate==0.7.5')
        ''')



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

from sos.actions import SoS_Action, SoS_ExecuteScript


@SoS_Action(acceptable_args=["script", "args"])
def ruby(script, args="", **kwargs):
    """Execute specified script using ruby. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script."""
    return SoS_ExecuteScript(script, "ruby", ".rb", args).run(**kwargs)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import copy
import os
import sys

import pytest
from sos import execute_workflow
from sos.eval import interpolate
from sos.parser import SoS_Script
from sos.targets import file_target, path, paths, sos_step, sos_targets
from sos.utils import env
from sos.workflow_executor import Base_Executor


def test_target_label():
    """Test labels of sos_targets"""
    a = sos_targets("a")
    assert a.labels == [""]
    b = sos_targets(["a", "b"])
    assert b.labels == ["", ""]
    c = sos_targets(["a1", "b1"], _source="here")
    assert c.labels == ["here", "here"]
    c.extend(b)
    assert c.labels == ["here", "here", "", ""]
    #
    assert c.select("").labels == ["", ""]
    assert c.select("here").labels == ["here", "here"]
    assert c["here"], [file_target("a1") == file_target("b1")]
    assert isinstance(c["here"], sos_targets)
    #
    # function item
    assert isinstance(c.select(1), sos_targets)
    assert c.select(1).labels == ["here"]
    assert c.select(1) == ["b1"]
    #
    # test slice of groups
    res = sos_targets(a=["a.txt", "b.txt"], b=["c.txt", "d.txt"], group_by=1)
    assert len(res.groups) == 4
    assert res.labels, ["a", "a", "b", "b"]
    res_a = res["a"]
    assert len(res_a) == 2
    assert res_a.labels == ["a", "a"]
    assert len(res_a.groups) == 4
    assert len(res_a.groups[0]) == 1
    assert len(res_a.groups[1]) == 1
    assert len(res_a.groups[2]) == 0
    assert len(res_a.groups[3]) == 0


def test_remove_targets():
    """Test sos_target.remove_targets()"""
    a = sos_targets(sos_step("1"), "a.txt")._group(by=1)
    a.remove_targets(type=sos_step)
    assert len(a) == 1
    assert len(a.groups) == 2
    assert len(a._groups[0]._indexes) == 0
    assert len(a._groups[0]._labels) == 0
    assert a._groups[1]._indexes == [0]
    assert len(a._groups[1]._labels) == 1


def test_sos_targets_signature(clear_now_and_after):
    """Test save and validate signatures of sos_targets"""
    clear_now_and_after('a.txt', 'b.txt')
    with open("a.txt", "w") as a:
        a.write("text1")
    with open("b.txt", "w") as b:
        b.write("text2")
    t = sos_targets("a.txt", "b.txt")
    sig = t.target_signature()
    assert t.validate(sig)
    # variables does not affect signature
    t[0].set("a", 2)
    assert sig == t.target_signature()
    #
    t.set("cc", "another string")
    assert t.validate(sig)


def test_target_set_get():
    """Test set and get attributes from targets"""
    a = file_target("a")
    a.set(b=1)
    assert a.b == 1
    assert a.get("b") == 1
    with pytest.raises(Exception):
        a.set("touch", 1)


def test_target_group_by():
    """Test new option group_by to sos_targets"""
    res = sos_targets(
        "e.txt",
        "f.ext",
        a=["a.txt", "b.txt"],
        b=["c.txt", "d.txt"],
        group_by=1)
    assert len(res.groups) == 6
    assert res.labels == ["", "", "a", "a", "b", "b"]
    #
    res = sos_targets(res, group_by=2)
    assert len(res.groups) == 3


def test_target_paired_with():
    """Test paired_with targets with vars"""
    res = sos_targets(
        "e.txt",
        "f.ext",
        a=["a.txt", "b.txt"],
        b=["c.txt", "d.txt"],
        group_by=1).paired_with("_name", ["e", "f", "a", "b", "c", "d"])
    for i, n in enumerate(["e", "f", "a", "b", "c", "d"]):
        assert res[i]._name == n
    #
    res = copy.deepcopy(res)
    for i, n in enumerate(["e", "f", "a", "b", "c", "d"]):
        assert res[i]._name == n
    #
    # test assert for length difference
    with pytest.raises(Exception):
        sos_targets("e.txt",
                    "f.ext").paired_with("name", ["e", "f", "a", "b", "c", "d"])


def test_target_group_with():
    """Test group_with targets with vars"""
    res = sos_targets(
        "e.txt",
        "f.ext",
        a=["a.txt", "b.txt"],
        b=["c.txt", "d.txt"],
        group_by=2).group_with("name", ["a1", "a2", "a3"])
    for i, n in enumerate(["a1", "a2", "a3"]):
        assert res.groups[i].name == n
    #
    res = copy.deepcopy(res)
    for i, n in enumerate(["a1", "a2", "a3"]):
        assert res.groups[i].name == n
    #
    # test assert for length difference
    with pytest.raises(Exception):

        sos_targets(
            "e.txt", "f.ext", group_by=1).group_with(
                "name",
                ["e", "f", "g"],
            )


def test_group_with_with_no_output():
    execute_workflow(r"""
[10]
input: for_each=dict(i=range(3))
output: group_with=dict(var=['A', 'B', 'C'][_index])
print(i)

[20]
print(f'Input is {_input} {var}')
""")


def test_merging_of_sos_targets():
    """Test merging of multiple sos targets"""
    # merge 0 to 0
    res = sos_targets("a.txt", "b.txt",
                      sos_targets("c.txt", "d.txt", group_by=1))
    assert len(res) == 4
    assert len(res.groups) == 2
    assert res.groups[0] == ["a.txt", "b.txt", "c.txt"]
    assert res.groups[1] == ["a.txt", "b.txt", "d.txt"]
    # merge N to N
    N1 = sos_targets("c.txt", "d.txt", group_by=1)
    N2 = sos_targets("a1.txt", "a2.txt", "a3.txt", "a4.txt", group_by=2)
    res = sos_targets(N1, N2)
    assert len(res) == 6
    assert len(res.groups) == 2
    assert res.groups[0] == ["c.txt", "a1.txt", "a2.txt"]
    assert res.groups[1] == ["d.txt", "a3.txt", "a4.txt"]
    # test N to M
    N1 = sos_targets("c.txt", "d.txt", group_by=1)
    N2 = sos_targets("a1.txt", "a2.txt", "a3.txt", "a4.txt", group_by=1)

    with pytest.raises(Exception):
        sos_targets(N1, N2)
    # merge 1 to N
    N1 = sos_targets("c.txt", "d.txt", group_by="all")
    N2 = sos_targets("a1.txt", "a2.txt", "a3.txt", "a4.txt", group_by=2)
    res = sos_targets(N1, N2)
    assert len(res) == 6
    assert len(res.groups) == 2
    assert res.groups[0] == ["c.txt", "d.txt", "a1.txt", "a2.txt"]
    assert res.groups[1] == ["c.txt", "d.txt", "a3.txt", "a4.txt"]
    # merge N to 1
    res = sos_targets(N2, N1)
    assert len(res) == 6
    assert len(res.groups) == 2
    assert res.groups[0] == ["a1.txt", "a2.txt", "c.txt", "d.txt"]
    assert res.groups[1] == ["a3.txt", "a4.txt", "c.txt", "d.txt"]


def test_target_format():
    """Test string interpolation of targets"""
    for target, fmt, res in [
        ("a.txt", "", "a.txt"),
        (sos_targets("a.txt"), "", "a.txt"),
        (sos_targets(["a.txt"]), "", "a.txt"),
        (sos_targets([r"c:\path\a.txt"]), "p", "/c/path/a.txt"),
        (sos_targets([r"c:path\a.txt"]), "p", "/c/path/a.txt"),
        (sos_targets("/a/b/a.txt"), "b", "a.txt"),
        (sos_targets("a b.txt"), "q", ("'a b.txt'", '"a b.txt"')),
        (sos_targets("a b.txt"), "x", ".txt"),
    ]:
        if isinstance(res, str):
            assert interpolate(
                "{{target:{}}}".format(fmt), globals(),
                locals()) == res, "Interpolation of {}:{} should be {}".format(
                    target, fmt, res)

        else:
            assert interpolate("{{target:{}}}".format(fmt), globals(), locals(
            )) in res, "Interpolation of {}:{} should be one of {}".format(
                target, fmt, res)


def test_iter_targets():
    """Test iterator interface of targets"""
    t = sos_targets("1", "2", ["3", "4"])
    assert len(t) == 4
    for idx, i in enumerate(t):
        assert str(i) == str(idx + 1)


def test_expand_wildcard():
    """test wildcard expansion of sos_targets"""
    a = sos_targets("*.py")
    assert len(a) > 1
    #
    a = paths("*.py")
    assert len(a) > 1


def test_shared():
    """Test option shared"""
    script = SoS_Script(r"""
parameter: res = 1

[0]
res = 2

[1]
res = 3
""")
    wf = script.workflow()
    Base_Executor(wf).run()
    assert env.sos_dict["res"] == 1
    #
    env.sos_dict.pop("res", None)
    script = SoS_Script(r"""
parameter: res = 1

[0: shared='res']
res = 2

[1]
res = 3
""")
    wf = script.workflow()
    Base_Executor(wf).run()
    assert env.sos_dict["res"] == 2
    #
    env.sos_dict.pop("res", None)
    script = SoS_Script(r"""
parameter: res = 1
parameter: a = 30

[0: shared='a']
res = 2

[1: shared='res']
res = 3
a = 5

""")
    wf = script.workflow()
    Base_Executor(wf).run()
    assert env.sos_dict["res"] == 3
    assert env.sos_dict["a"] == 30
    # test multiple vars
    env.sos_dict.pop("res", None)
    script = SoS_Script(r"""
parameter: res = 1
parameter: a = 30

[1: shared=('res', 'a')]
res = 3
a = 5

""")
    wf = script.workflow()
    Base_Executor(wf).run()
    assert env.sos_dict["res"] == 3
    assert env.sos_dict["a"] == 5
    #
    # test expression
    env.sos_dict.pop("res", None)
    script = SoS_Script(r"""
parameter: res = 1
parameter: a = 30

[1: shared={'res': 'res + 6', 'c': 'a'}]
res = 3
a = 5

""")
    wf = script.workflow()
    Base_Executor(wf).run()
    assert env.sos_dict["res"] == 9
    assert env.sos_dict["c"] == 5
    # test mixed vars and mapping
    env.sos_dict.pop("res", None)
    script = SoS_Script(r"""
parameter: res = 1
parameter: a = 30

[1: shared=['res', {'c': 'a'}]]
res = 3
a = 5

""")
    wf = script.workflow()
    Base_Executor(wf).run()
    assert env.sos_dict["res"] == 3
    assert env.sos_dict["c"] == 5
    # test the step_ version of variables
    script = SoS_Script(r"""
parameter: res = 1
parameter: a = 30

[1: shared=['res', {'c': 'sum(step_a)'}]]
input: for_each={'i': range(10)}
a = _index**2

""")
    wf = script.workflow()
    Base_Executor(wf).run()
    assert env.sos_dict["c"] == sum(x**2 for x in range(10))


#    def testSectionOptionWorkdir():
#        '''Test section option workdir'''
#        script = SoS_Script(r"""
#
# [1: workdir='tmp']
# run:
#    touch 'a.txt'
# """)
#        wf = script.workflow()
#        Base_Executor(wf).run()
#        assert os.path.isfile('tmp/a.txt'))
#        shutil.rmtree('tmp')


def test_depends_executable():
    """Testing target executable."""
    script = SoS_Script("""
[0]
depends: executable('python --version', '3.')
run:
touch a.txt
""")
    wf = script.workflow()
    if file_target("a.txt").exists():
        file_target("a.txt").unlink()
    Base_Executor(wf).run()
    assert os.path.isfile("a.txt")
    file_target("a.txt").unlink()


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows executable cannot be created with chmod.")
def test_output_executable(clear_now_and_after):
    """Testing target executable."""
    # change $PATH so that lls can be found at the current
    # directory.
    clear_now_and_after('lls')
    os.environ["PATH"] += os.pathsep + "."
    script = SoS_Script("""
[0]
output: executable('lls')
run:
touch lls
chmod +x lls
""")
    wf = script.workflow()
    if file_target("lls").exists():
        file_target("lls").unlink()
    env.config["sig_mode"] = "force"
    Base_Executor(wf).run()
    # test validation
    env.config["sig_mode"] = "default"
    Base_Executor(wf).run()
    if file_target("ls").exists():
        file_target("lls").unlink()


def test_depends_env_variable():
    """Testing target env_variable."""
    if file_target("a.txt").exists():
        file_target("a.txt").unlink()
    if sys.platform == "win32":
        script = SoS_Script("""
[0]
depends: env_variable('AA')
output:  'a.txt'
run:
echo %AA% > a.txt
""")
    else:
        script = SoS_Script("""
[0]
depends: env_variable('AA')
output:  'a.txt'
run:
echo $AA > a.txt
""")
    wf = script.workflow()
    os.environ["AA"] = "A1"
    Base_Executor(wf).run()
    with open("a.txt") as at:
        assert at.read().strip() == "A1"
    # test validation
    Base_Executor(wf).run()
    # now if we change var, it should be rerun
    os.environ["AA"] = "A2"
    Base_Executor(wf).run()
    # allow one second variation
    with open("a.txt") as at:
        assert at.read().strip() == "A2"
    file_target("a.txt").unlink()


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows executable cannot be created with chmod.")
def test_provides_executable():
    """Testing provides executable target."""
    # change $PATH so that lls can be found at the current
    # directory.
    os.environ["PATH"] += os.pathsep + "."
    if file_target("lls").exists():
        file_target("lls").unlink()
    script = SoS_Script("""
[lls: provides=executable('lkls')]
run:
touch lkls
chmod +x lkls

[c]
depends: executable('lkls')

""")
    wf = script.workflow("c")
    Base_Executor(wf).run()
    file_target("lkls").unlink()


def test_shared_var_in_paired_with(temp_factory):
    temp_factory(["1.txt", "2.txt"])
    for file in ("1.out", "2.out", "1.out2", "2.out2"):
        if file_target(file).exists():
            file_target(file).unlink()
    script = SoS_Script("""
[work_1: shared = {'data': 'step_output'}]
input: "1.txt", "2.txt", group_by = 'single', pattern = '{name}.{ext}'
output: expand_pattern('{_name}.out')
run: expand=True
touch {_output}

[work_2]
depends: sos_variable('data')
input: "1.txt", "2.txt", group_by = 'single', pattern = '{name}.{ext}', paired_with = dict(data1=data)
output: expand_pattern('{_name}.out2')
run: expand=True
touch {data1[0]} {_output}
""")
    wf = script.workflow()
    Base_Executor(wf).run()
    for file in ("1.out", "2.out", "1.out2", "2.out2"):
        if file_target(file).exists():
            file_target(file).unlink()


def test_shared_var_in_for_each(temp_factory, clear_now_and_after):
    temp_factory("1.txt", "2.txt")
    clear_now_and_after("1.out", "2.out", "1.out2", "2.out2", '2.out_2.out2', '1.out_1.out2', '1.out_2.out2', '2.out_1.out2')
    script = SoS_Script("""
[work_1: shared = {'data': 'step_output'}]
input: "1.txt", "2.txt", group_by = 'single', pattern = '{name}.{ext}'
output: expand_pattern('{_name}.out')
run: expand=True
touch {_output}

[work_2]
depends: sos_variable('data')
input: "1.txt", "2.txt", group_by = 'single', for_each = dict(data=data),  pattern = '{name}.{ext}'
output: expand_pattern('{data}_{_name}.out2')
run: expand=True
touch {_output}

""")
    wf = script.workflow()
    Base_Executor(wf).run()


def test_removed_depends(clear_now_and_after):
    """Test a case where a dependent file has signature, but
    gets removed before the next run."""
    clear_now_and_after('a.txt', 'b.txt')
    script = """
        [tet: provides='a.txt']
        run:
        echo "something" > a.txt

        [20]
        depends: 'a.txt'
        output: 'b.txt'
        run:
        cat a.txt > b.txt
        """
    # this should be ok.
    execute_workflow(script)
    # now let us remove a.txt (but the signature is still there)
    clear_now_and_after('a.txt', 'b.txt')
    execute_workflow(script)


def test_sos_step(clear_now_and_after):
    """Test target sos_step"""
    clear_now_and_after("t1.txt", "t2.txt", "5.txt", "10.txt", "20.txt")

    script = """
        [t1]
        run:
        touch t1.txt

        [t2: provides='t2.txt']
        depends: sos_step('t1')
        run:
        touch t2.txt

        [5]
        run:
        touch 5.txt

        [10]
        depends: sos_step('t2')
        run:
        touch 10.txt

        [20]
        depends: sos_step('t1')
        run:
        touch 20.txt
        """
    # this should be ok.
    execute_workflow(script, options={'sig_mode': 'force'})
    for file in ["t1.txt", "t2.txt", "5.txt", "10.txt", "20.txt"]:
        assert file_target(file).target_exists(), file + " should exist"


def test_zap(clear_now_and_after):
    """Test zap"""
    clear_now_and_after('testzap1.txt.zapped', 'testzap.txt.zapped')
    with open("testzap.txt", "w") as sf:
        sf.write("some text")
    path("testzap.txt").zap()
    assert os.path.isfile("testzap.txt.zapped")
    assert not os.path.isfile("testzap.txt")
    # re-zap is ok
    file_target("testzap.txt").zap()
    assert os.path.isfile("testzap.txt.zapped")
    assert not os.path.isfile("testzap.txt")
    # non-existent file
    os.remove("testzap.txt.zapped")
    with pytest.raises(FileNotFoundError):
        path("testzap.txt").zap()
    #
    with open("testzap.txt", "w") as sf:
        sf.write("some text")
    with open("testzap1.txt", "w") as sf:
        sf.write("some text")
    paths("testzap.txt", "testzap1.txt").zap()
    assert os.path.isfile("testzap.txt.zapped")
    assert not os.path.isfile("testzap.txt")
    assert os.path.isfile("testzap1.txt.zapped")
    assert not os.path.isfile("testzap1.txt")
    #
    os.remove("testzap.txt.zapped")
    os.remove("testzap1.txt.zapped")
    with open("testzap.txt", "w") as sf:
        sf.write("some text")
    with open("testzap1.txt", "w") as sf:
        sf.write("some text")
    sos_targets(["testzap.txt", "testzap1.txt"]).zap()
    assert os.path.isfile("testzap.txt.zapped")
    assert not os.path.isfile("testzap.txt")
    assert os.path.isfile("testzap1.txt.zapped")
    assert not os.path.isfile("testzap1.txt")


def test_zap_run(clear_now_and_after):
    """Test run with zapped input files"""
    clear_now_and_after('zap1.txt.zapped')
    with open("zap1.txt", "w") as sf:
        sf.write("seomething")
    script = """\
        [1]
        input: 'zap1.txt'
        output: "zap2.txt"
        run:
            echo asd > zap2.txt
        _input.zap()
        """
    execute_workflow(script, options={"sig_mode": "force"})

    assert os.path.isfile("zap1.txt.zapped")
    assert not os.path.isfile("zap1.txt")
    assert os.path.isfile("zap2.txt")
    # can run again
    env.config["sig_mode"] = "default"

    execute_workflow(script)
    # now if we remove target
    os.remove("zap2.txt")
    with pytest.raises(Exception):
        execute_workflow(script)


def test_system_resource():
    """Test targtet system_resource"""
    execute_workflow("""\
        [1: shared='a']
        depends: system_resource(mem='1M',disk='1M')
        a = 1
        """)
    assert env.sos_dict["a"] == 1

    #
    with pytest.raises(Exception):
        execute_workflow("""\
            [1: shared='a']
            depends: system_resource(mem='1T')
            a = 1
            """)

    with pytest.raises(Exception):
        execute_workflow("""\
            [1: shared='a']
            depends: system_resource(disk='10P')
            a = 1
        """)


def test_temp_file():
    """Test sos_tempfile target"""
    execute_workflow("""
        filename = sos_tempfile(name='a')

        assert filename == sos_tempfile(name='a')
        assert filename != sos_tempfile(name='b')
        assert filename != sos_tempfile()
        assert sos_tempfile() != sos_tempfile()

        assert sos_tempfile(prefix='some') != sos_tempfile(prefix='some')
        assert sos_tempfile(suffix='.sh') != sos_tempfile(suffix='.sh')
        assert sos_tempfile(dir='.') != sos_tempfile(dir='.')

        assert sos_tempfile('a.txt') == sos_tempfile('a.txt')
        assert sos_tempfile(path('a.txt')) == sos_tempfile(path('a.txt'))
        assert sos_tempfile(file_target('a.txt')) == sos_tempfile(file_target('a.txt'))

    """)


def test_named_path():
    """Test the use of option name of path"""
    execute_workflow(
        """
        import os
        # windows might not have HOME
        if 'HOME' in os.environ:
            assert path('#home') == os.environ['HOME']
            assert 'home' in path.names()
            assert 'home' in path.names('docker')
        """,
        options={
            "config_file": os.path.join(os.path.expanduser("~"), "docker.yml")
        },
    )


@pytest.mark.skipif(
    sys.platform == 'win32', reason='Graphviz not available under windows')
def test_to_named_path_path():
    execute_workflow(
        """
        [10: shared="a"]
        a = path('/root/xxx/whatever').to_named_path(host='docker')
        """,
        options={
            "config_file": os.path.join(os.path.expanduser("~"), "docker.yml"),
        })
    assert env.sos_dict['a'] == '#home/xxx/whatever'
